import unittest

from dsframework.base.server.cache.cache import CacheItem, Cache


class TestCacheItem(unittest.TestCase):
    def test_cache_item(self):
        cache_item = CacheItem("test_key", "test_value")
        self.assertEqual(cache_item.key, "test_key")
        self.assertEqual(cache_item.value, "test_value")


class TestCache(unittest.TestCase):
    def test_cache_abstract_methods(self):
        cache = Cache()
        with self.assertRaises(NotImplementedError):
            cache.get([])
        with self.assertRaises(NotImplementedError):
            cache.set([])
        with self.assertRaises(NotImplementedError):
            cache.status()



from typing import Any, Optional, Union, List
from dsframework.base.trainer.pl_wrapper import *


class ZIDS_PLModel(pl.LightningModule):
    """! Trainer base class inherits from LightningModule, which organizes your code into 6 sections:

        - Computations (init).
        - Train Loop (training_step)
        - Validation Loop (validation_step)
        - Test Loop(test_step)
        - Prediction Loop (predict_step)
        - Optimizers and LR Schedulers(configure_optimizers)

        Log parameters:
        - For single value, use:
            self.log("test_loss", test_loss)  # prog_bar=True

        - For multiple values use dictionary, use:
            values = {"loss": loss, "acc": acc, "metric_n": metric_n}  # add more items if needed
            self.log_dict(values)

        View results in tensorboard:
            tensorboard --logdir=lightning_logs/

        Loading datasets:

            example:
            @code{.py}
            train_set = datasets.MNIST(os.getcwd() + '/data', download=True, train=True, transform=transform)
            test_set = datasets.MNIST(os.getcwd() + '/data', download=True, train=False, transform=transform)

            # use 20% of training data for validation
            train_set_size = int(len(train_set) * 0.8)
            valid_set_size = len(train_set) - train_set_size

            # split the train set into two
            seed = torch.Generator().manual_seed(42)
            train_set, valid_set = random_split(train_set, [train_set_size, valid_set_size], generator=seed)

            train_set_loader = DataLoader(train_set, num_workers=2)
            val_set_loader = DataLoader(valid_set, num_workers=2)
            test_set_loader = DataLoader(test_set, num_workers=2)
            @endcode


        Training:
            example:
            @code{.py}
            trainer = pl.Trainer(max_epochs=1, callbacks=[checkpoint_callback])
            trainer.fit(model=autoencoder, train_dataloaders=train_set_loader, val_dataloaders=val_set_loader)
            @endcode

        Testing:
            example:
            @code{.py}
            trainer.test(model=autoencoder, dataloaders=test_set_loader)
            @endcode

        Loading a trained model, use:
            - load_from_checkpoint

                example:
                @code{.py}
                model = MyTrainer.load_from_checkpoint('lightning_logs/epoch=0-step=48000.ckpt')
                @endcode


        Saving a model, use:
            - save_checkpoint

                for example:
                @code{.py}
                trainer.save_checkpoint("my_checkpoint.ckpt")
                @endcode

            - ModelCheckpoint - define automated checkpoint saving, use:
                for example:
                @code{.py}
                checkpoint_callback = ModelCheckpoint(dirpath="lightning_logs/", save_top_k=2, monitor="val_loss")
                @endcode


    """

    def __init__(self, model_config=None, metrics_config=None, trainer_config=None):
        super().__init__()
        self.train_accuracy = None
        self.val_accuracy = None
        self.test_accuracy = None

        self.train_precision = None
        self.val_precision = None
        self.test_precision = None

        self.train_recall = None
        self.val_recall = None
        self.test_recall = None

        self.train_f1 = None
        self.val_f1 = None
        self.test_f1 = None

        self.train_confusion_matrix = None
        self.val_confusion_matrix = None
        self.test_confusion_matrix = None

        self.model_config = model_config
        self.trainer_config = trainer_config
        self.metrics = metrics_config
        self.loss_function = self.model_config.get('loss_function', None)
        self.test_results = None

    def configure_optimizers(self):
        """! Choose what optimizers and learning-rate schedulers to use in your optimization. Normally you’d need one.
        But in the case of GANs or similar you might have multiple.

            Returns: Any of this 6 options
                - Single optimizer.
                - List or Tuple of optimizers.
                - Two lists - The first list has multiple optimizers, and the second has multiple LR schedulers
                    (or multiple lr_scheduler_config).
                - Dictionary, with an "optimizer" key, and (optionally) a "lr_scheduler" key whose value
                    is a single LR scheduler or lr_scheduler_config.
                - Tuple of dictionaries as described above, with an optional "frequency" key.
                - None - Fit will run without any optimizer.

        Example:
            @code{.py}
            optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
            return optimizer
            @endcode

        Another example:
            @code{.py}
            SGD_optimizer = optim.SGD(model.parameters(), lr=0.001, momentum=0.7)
            return SGD_optimizer
            @endcode

        """
        pass

    def forward(self, *args, **kwargs) -> Any:
        """! Defines the computation performed at every call.

        Implementation example:
            @code{.py}
            return self.model(x)
            @endcode
        """
        pass

    def training_step(self, batch, batch_idx) -> STEP_OUTPUT:
        """! Override to enable training loop

        Implementation example:
            @code{.py}
            def training_step(self, batch, batch_idx):
                x, y = batch
                x = x.view(x.size(0), -1)
                z = self.encoder(x)
                x_hat = self.decoder(z)
                loss = F.mse_loss(x_hat, x)
                return loss
            @endcode

        If you need to do something with all the outputs of each training_step(), override the
        training_epoch_end() method.
        """
        x, y = batch
        y_hat = self.network(x)
        loss = self.loss_function(y_hat, y)
        self.log("train_loss", loss, prog_bar=True, logger=True)

        return loss

    def validation_step(self, batch, batch_idx) -> Optional[STEP_OUTPUT]:
        """! Override to enable validation loop

        Implementation example:
            @code{.py}
            def validation_step(self, batch, batch_idx):
                x, y = batch
                x = x.view(x.size(0), -1)
                z = self.encoder(x)
                x_hat = self.decoder(z)
                test_loss = F.mse_loss(x_hat, x)
                self.log("val_loss", test_loss, prog_bar=True)
            @endcode
        """
        x, y = batch
        y_hat = self.network(x)
        loss = self.loss_function(y_hat, y)

        self.log("val_loss", loss)

        return loss

    def test_step(self, batch, batch_idx) -> Optional[STEP_OUTPUT]:
        """! Override to enable test loop.

        Implementation example:
            @code{.py}
            x, y = batch
            x = x.view(x.size(0), -1)
            z = self.encoder(x)
            x_hat = self.decoder(z)
            test_loss = F.mse_loss(x_hat, x)
            self.log("test_loss", test_loss)  # prog_bar=True

            return {'loss': loss, 'preds': pred, 'target': y}  # will be available in test_epoch_end
            @endcode
        """
        x, y = batch
        y_hat = self.network(x)
        loss = self.loss_function(y_hat, y)
        self.log("test_loss", loss)

        return loss

    def training_epoch_end(self, outputs: EPOCH_OUTPUT) -> None:
        """! Override method to implement functionality related to training_step() outputs.

        training step accumulated returns:
            @code
            all_preds = torch.stack(training_step_outputs)
            @endcode
        """
        pass

    def validation_epoch_end(self, outputs: Union[EPOCH_OUTPUT, List[EPOCH_OUTPUT]]) -> None:
        """! Override method to implement functionality related to validation_step() outputs."""
        pass

    def training_step_end(self, step_output: STEP_OUTPUT) -> STEP_OUTPUT:
        """! When the data is split across multiple GPUs, this method will have outputs from all devices that can be
        accumulated to get the effective results.
        """
        pass

    def validation_step_end(self, *args, **kwargs) -> Optional[STEP_OUTPUT]:
        """! When the data is split across multiple GPUs, this method will have outputs from all devices that can be
        accumulated to get the effective results.
        """
        pass

    def test_epoch_end(self, outputs):
        """! Implement this to calculate test results and fill the local dictionary return self.test_results
        Args:
            outputs: Array (ie. dict, list etc.) with accumulated results returned from:
             def test_step(self, batch, batch_idx)


        Results structure example:
            self.test_results = {
                'precision': precision_result,
                'accuracy': accuracy_result,
                'recall': recall_result,
                'f1': f1_result,
                'confusion_matrix': conf_matrix_results}

        Confusion matrix example:
            preds = torch.cat([tmp['preds'] for tmp in outputs])
            targets = torch.cat([tmp['target'] for tmp in outputs])

            conf_matrix_results = self.confusion_matrix(preds, targets)
        """
        pass

    def get_test_results(self):
        return self.test_results

    def log_metrics(self, pred, y, step_name):
        pass

    def prepare_test_results(self):
        pass



import json
from typing import List, Optional

from pydantic import validator, Extra
from pydantic.main import BaseModel

##
# @file
# @brief Outputs from the DSF -> DSP

class TestOutputResponse(BaseModel):
    """! BaseModel (pydantic.main) the base class for TestOutputResponse.
    @verbatim
    Args:
        truth_id: int
        truth_dataset_id(int): dataset id from the portal
        model_type_id : Model ID as defined by a table held in staging DSP (number per project)
        raw_id(=-1) : int
        pred(str): prediction, take all input data, replace ground truth with actual model prediction, and store as a json string
        target(str): the target, take all the data as a string (from the input dataset)
        text: string
    @endverbatim
    """
    truth_id: int
    truth_dataset_id: int
    model_type_id: int
    raw_id: int = -1
    pred: str
    target: str
    text: str = ''

    @validator('pred')
    def pred_must_be_valid_json(cls, v):
        """! Method that check the data if it's valid json string.

            Args:
                cls : not in use
                v : Data that is going to be checked.
            Raises:
                ValueError
        """
        try:
            json.loads(v)
        except:
            raise ValueError('pred must be a valid json string')
        return v

    @validator('target')
    def target_must_be_valid_json(cls, v):
        """! Method that check the data if it's valid json string.

            Args:
                cls : not in use
                v : Data that is going to be checked.
            Raises:
                ValueError
        """
        try:
            json.loads(v)
        except:
            raise ValueError('target must be a valid json string')
        return v

    class Config:
        extra = Extra.allow



from dsframework.base.trainer.test import ZIDSTrainerTest
from trainer.pl_wrapper.plmodel import generatedProjectNamePlmodel
from trainer.data import generatedProjectNameData
from dsframework.base.trainer.pl_wrapper import *

##
# @file
# @brief generatedClass class, executes test process.


class generatedClass(ZIDSTrainerTest):
    """! Test class, runs the model against a labeled test dataset to evaluate its metrics.

        Three things needs to be done for this to work:
            1. Receive data_module with labeled dataset loaded.
            2. Load model
            3. Create and Run trainer instance.
    """

    def __init__(self, model_path=None, test_dataset_path=None, trainer_config=None, model_config=None, metrics_config=None,
                 save_path=''):
        """generatedClass Initializer.

        Runs trainer_setup() and load_model() methods/
        """

        self.data_class_test = None
        self.model_config = model_config
        self.metrics_config = metrics_config
        self.save_path = save_path

        super().__init__(model_path, test_dataset_path, trainer_config)

    def load_model(self, model_path):
        """! Loads a model

        Needs to be implemented.

        Example (pytorch lightning):
            self.model = generatedProjectNamePlmodel.load_from_checkpoint(
                model_path,
                model_config=self.model_config,
                metrics_config=self.metrics_config,
                trainer_config=self.trainer_config,
                save_path=self.save_path
            )
        """

        if self.model is None:
            raise Exception('load_model() method in test.py, not implement.')

    def create_data_module(self, tokenizer, test_dataset_path):
        """! Create the data module to use with the test process.


        Example:
            self.data_class_test = generatedProjectNameData(
                model_config=self.model_config,
                load_datasets=True,
                test_dataset_path=self.test_dataset_path,
                tokenizer=self.model.tokenizer)
        """
        if self.data_class_test is None:
            raise Exception('create_data_module() method in test.py, not implement.')

    def trainer_setup(self, trainer_config):
        """! Instantiate trainer class

        Needs to be implemented.

        Example (pytorch lightning):
            self.trainer = pl.Trainer(**trainer_config)
        """

        if self.trainer is None:
            raise Exception('trainer_setup() method in test.py, not implement.')

    def execute_test(self):
        """! Implement test process and return the results.

        Returns:
            test results - Metric results from test.

        Example (pytorch lightning):
            if self.model is None or self.data_class_test is None or self.trainer is None:
                return None
            self.trainer.test(self.model, self.data_class_test.get_data_module())

            return self.model.get_test_results()
        """

        raise Exception('Methods execute_test() not implemented in test.py.')



import subprocess
import os
import sys
from ujson import dumps as json_dumps
from ujson import load as json_load
from ujson import loads as json_loads
from yaml import dump as yaml_dump

from datetime import timedelta
from datetime import datetime

from google.cloud.dataproc_v1 import ClusterControllerClient
from google.cloud.dataproc_v1 import WorkflowTemplatePlacement
from google.cloud.dataproc_v1.types import OrderedJob
from google.cloud.dataproc_v1.types import PySparkJob
from google.cloud.dataproc_v1.services.workflow_template_service.async_client import WorkflowTemplateServiceClient
from google.cloud.dataproc_v1.types.workflow_templates import WorkflowTemplate


##
# @file
# @brief Defines workflow base class.
class ZIDS_Workflow():
    def __init__(self, workflow_name: str = ""):
        """! ZIDS_Workflow initializer.
        Loads config file and initializes parameters."""

        self.wf_name = workflow_name

        file_path = sys.modules[self.__class__.__module__].__file__
        self.curr_path = file_path[:file_path.rfind("batch_files")]

        with open(self.curr_path + 'batch_files/batch_config.json') as batch_cfg_file:
            batch_config = json_load(batch_cfg_file)

        with open(self.curr_path + 'batch_files/workflows/' + self.wf_name +
                  '_workflow/config_' + self.wf_name + '.json') as wf_cfg_file:
            wf_config = json_load(wf_cfg_file)

        self.config = {**batch_config, **wf_config}

        self.user_email = ''
        env_type = os.environ.get('SPRING_PROFILES_ACTIVE')
        if (not env_type == 'production') and (not env_type == 'staging'):
            command = 'gcloud config get-value account'
            user_email = subprocess.check_output(command, shell=True).decode(sys.stdout.encoding)
            self.user_email = '/' + user_email.split("@")[0]

        if self.user_email:
            self.config['user_email'] = self.user_email

        self.default_args = [json_dumps(self.config),
                             "start_date_placeholder", "end_date_placeholder", "params_placeholder"]

        self.region = self.config['region']
        self.zone = self.region + '-b'
        self.project_id = self.config['project_id']
        self.template_id = self.config['template_id']
        self.bucket_name = self.config['bucket_name']
        self.project_name = self.config['project_name']
        self.autoscaling_policy_id = self.project_name + '-auto-scaling-policy'
        self.wf_stages = self.config['workflow_stages']

        self.unique_iteration_id = self.config['unique_iteration_id']
        self.unique_template_id = self.config['unique_template_id']
        if self.unique_template_id:
            self.template_id = self.template_id + '_' + self.unique_template_id

        if self.unique_iteration_id:
            self.folder_path = self.project_name + self.user_email + \
                               '/' + self.wf_name + '/unique_iteration_id_' + self.unique_iteration_id
        else:
            self.folder_path = self.project_name + self.user_email + '/' + self.wf_name + '/main'

        self.bucket_path = f'gs://{self.bucket_name}/{self.folder_path}'
        self.project_path = 'projects/{project_id}/regions/{region}'.format(project_id=self.project_id,
                                                                            region=self.region)

        if self.config['cluster_conf']['managed_cluster']:

            # Configure initialization script
            cluster_init_action = {"executable_file": f"{self.bucket_path}"
                                                      f"/batch_files/scripts/cluster-init-actions.sh"}
            cluster_init_list = self.config['cluster_conf']['managed_cluster']['config']['initialization_actions']

            if not isinstance(cluster_init_list, list):
                cluster_init_list = []

            cluster_init_list.append(cluster_init_action)

            # Take general parameters regarding GCE configuration from the main config file, as they can be modified
            # by Jenkins when deploying the service
            master_gce_config = batch_config['cluster_conf']['managed_cluster']['config']['gce_cluster_config']
            self.config['cluster_conf']['managed_cluster']['config']['gce_cluster_config'] = master_gce_config

        # Define workflow client
        self.serv_client = WorkflowTemplateServiceClient(
            client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"}
        )

        # Define extra stage parameters
        self.extra_params = json_dumps(self.config['extra_job_params'])
        # Selection between current date or batch_config.json as date
        self.instantiate_now = False if "date_str" in self.config else True

    def configure_pip_packages(self, requirements_path: str, append: bool = False):
        """Configure pip packages based on the required steps.

        Args:
            requirements_path: Path to requirements.txt file.
            append: Append requirement packages to the currently loaded ones.

        """

        with open(requirements_path) as req_file:
            required_pkg = req_file.read().splitlines()
            required_pkg = [x for x in required_pkg if "#" not in x]
            pip_packages = ' '.join(required_pkg)

        if self.config['cluster_conf']['managed_cluster']:
            # Configure pip packages
            if not append:
                self.config['cluster_conf']['managed_cluster']['config']['gce_cluster_config']['metadata'][
                    'PIP_PACKAGES'] \
                    = pip_packages
            else:
                self.config['cluster_conf']['managed_cluster']['config']['gce_cluster_config']['metadata'][
                    'PIP_PACKAGES'] \
                    += ' ' + pip_packages

    def check_cluster(self):
        """! ZIDS_Workflow get_cluster.
        Tries to locate an existing cluster. If exists, we'll use it for template instantiation.
        If not, we will use the managed cluster to create it based on configuration
        When clusters are created manually, they include the workflow name, so this function looks for it.
        If it was created automatically, then the name is generated and this function will not find it.
        """
        clusterClient = ClusterControllerClient(
            client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"}
        )

        cluster_name = f"{self.config['cluster_conf']['managed_cluster']['cluster_name']}-{self.wf_name}"

        try:
            exist_cluster = clusterClient.get_cluster(
                project_id=self.project_id,
                region=self.region,
                cluster_name=cluster_name,
            )
            if exist_cluster:
                del self.config['cluster_conf']['managed_cluster']
                self.config['cluster_conf']['cluster_selector'] = {
                    "cluster_labels": {
                        "goog-dataproc-cluster-name": cluster_name
                    }
                }
        except Exception as e:
            if "404" in str(e):
                print(f"Cluster {cluster_name} not found. Will be created when workflow is instantiated")
            else:
                # If there's something other than 404, we want to know, but the flow can carry on.
                print(f'Error getting cluster: {e}. Cluster will be created on demand')

    def build_template_structure(self) -> WorkflowTemplate:
        """! ZIDS_Workflow build_template_structure.
        build a workflow template class instance, containing all available configurations, mainly the managed cluster
        configuration at this point.

            Returns:
                WorkflowTemplate() - Created template with basic arguments
        """
        cluster_config = WorkflowTemplatePlacement(self.config['cluster_conf'])

        # if results not exist
        template = WorkflowTemplate()
        template.id = self.template_id
        template.name = self.project_path + '/workflowTemplates/' + template.id
        template.placement = cluster_config
        return template

    def create_ordered_job(self, stage_dir: str, spark_properties, python_files, archive_files,
                           additional_args=[], additional_jars=[]) -> OrderedJob:
        """! ZIDS_Workflow create_ordered_job.
        Create an ordered job from a stage directory

            Args:
                stage_dir: Stage directory, should hold the main.py file and configuration
                           example: batch_files/stages/stage_write_page_list_to_big_query
                spark_properties: A list, containing different spark properties
                python_files: A list of files which hold the project code (Can be archived)
                archive_files: A list of compressed files that the stage requires
                additional_args: A list of additional arguments, to add on top of the default ones
                additional_jars: A list of additional jar files, to add on top of the default ones
            Returns:
                An ordered job
        """

        # Assume that the last part of the path is the stage name
        stage_name = stage_dir.rsplit('/', 1)[-1]

        created_stage = OrderedJob()
        pyJob = PySparkJob()
        pyJob.main_python_file_uri = f'{self.bucket_path}/{stage_dir}/main.py'
        pyJob.archive_uris = [f'{self.bucket_path}/dist/{file_name}' if file_name.find('gs:') == -1
                              else file_name for file_name in archive_files]
        pyJob.python_file_uris = [f'{self.bucket_path}/dist/{file_name}' if file_name.find('gs:') == -1
                                  else file_name for file_name in python_files]
        pyJob.properties = spark_properties
        pyJob.args = self.default_args + additional_args
        pyJob.jar_file_uris = self.default_jars + additional_jars
        created_stage.step_id = f'{self.project_name}-{stage_name}'
        created_stage.pyspark_job = pyJob
        return created_stage

    def check_template(self):
        """! Check for an existing template. If found, delete it.
        """

        exist_template = self.get_workflow_template()

        # Deleting template if it exists
        if exist_template:
            self.delete_workflow_template(exist_template)

    def create_template(self, template):
        """! Create the complete template on the remote location.
            Args:
                template: WorkflowTemplate()
        """
        try:
            result = self.serv_client.create_workflow_template(
                parent=self.project_path,
                template=template
            )
            print(f'template {template.id} created successfully')
        except Exception as e:
            print(f'Error creating template: {e}')

    def get_workflow_template(self) -> WorkflowTemplate:
        """! Get the current template
            Returns:
                WorkflowTemplate, if such was retrieved

        """

        template_name = self.project_path + '/workflowTemplates/' + self.template_id

        try:
            existing_template = self.serv_client.get_workflow_template(
                name=template_name,
            )
            print(f'template {self.template_id} retrieved successfully')
        except Exception as e:
            print(f'Cannot get template, exception: {e}')
            return None

        return existing_template

    def instantiate_template(self, template):
        """! Instantiate the template into a working workflow, overriding DAG
        Do this only on dev mode
            Args:
                template: WorkflowTemplate()
        """

        if self.instantiate_now:
            start_date = datetime.now().replace(second=0, microsecond=0)
            end_date = start_date + timedelta(days=1)
            start_date_str = start_date.strftime("%Y-%m-%dT%H:%M:%S+00:00")
            end_date_str = end_date.strftime("%Y-%m-%dT%H:%M:%S+00:00")
        else:
            start_date_str = self.config['date_str']
            end_date_str = self.config['date_end']

        try:
            self.serv_client.instantiate_workflow_template(
                name=template.name,
                parameters={"START_DATE": start_date_str, "END_DATE": end_date_str, "PARAMS": self.extra_params}
            )
            print(f'template {template.id} instantiated successfully')
        except Exception as e:
            print(f'Error instantiating template: {e}')

    def delete_workflow_template(self, template):
        """! Delete a given template
            Args:
                template: WorkflowTemplate()
        """

        try:
            self.serv_client.delete_workflow_template(
                name=template.name,
            )
            print(f'template {template.id} deleted successfully')
        except Exception as e:
            print(f'Error deleting template: {e}')

    def create_workflow_template(self):
        """! Create the workflow template, along with all relevant stages
        """
        raise NotImplementedError

    def create_yaml_file(self, workflow_template):
        meta_msg = workflow_template._meta.parent
        workflow_str = meta_msg.to_json(workflow_template)
        clean_workflow_str = workflow_str.strip()
        workflow_json = json_loads(clean_workflow_str)
        # Remove unnecessary fields which are enums that are messing up yaml parsing. Needs to be resolved later if
        # they are needed. Import the Enum and give correct values
        del workflow_json['placement']['managedCluster']['config']['gceClusterConfig']['privateIpv6GoogleAccess']
        del workflow_json['placement']['managedCluster']['config']['masterConfig']['preemptibility']
        del workflow_json['placement']['managedCluster']['config']['workerConfig']['preemptibility']

        file_path = sys.modules[self.__class__.__module__].__file__
        curr_path = file_path[:file_path.rfind("workflow")]
        with open(curr_path + f'/workflow_{self.wf_name}.yaml', 'w') as yml:
            yaml_dump(workflow_json, yml)



import ast
from typing import List

import click
import os
import shutil
import re
import platform
import subprocess
import site
import webbrowser
import hashlib
from datetime import datetime
from pathlib import Path
from ujson import load as json_load
from ujson import dumps as json_dumps
from ujson import dump as json_dump
from dsframework.documentation.scripts import doxygen_handler

isWindows = platform.system() == 'Windows'

__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))

__proj_root__ = os.path.abspath(os.getcwd())

directories = {
    'main': '',
    'pipeline': '',
    'artifacts': '',
    'models': '',
    'vocabs': '',
    'other': '',
    'preprocessor': '',
    'predictables': '',
    'predictors': '',
    'forcers': '',
    'postprocessor': '',
    'schema': '',
    'tester': '',
    'test_schema': '',
    'trainer': '',
    'server': '',
    'dsp': '',
    'documentation': ''
}


##
# @file
# @brief This file defines and implements the framework CLI. It helps us to:\n
#        Create: Project/Forcer/Predictable/Tester-files/Deploy-files/Cloud-Eval-Files  \n
#        Run   : Server / Evaluation (CSV,Cloud) \n


class AliasedGroup(click.Group):
    """! AliasedGroup implements click.Group base class."""

    ##
    # @hidecallgraph @hidecallergraph
    def get_command(self, ctx, cmd_name):
        try:
            cmd_name = ALIASES[cmd_name].name
        except KeyError:
            pass
        return super().get_command(ctx, cmd_name)


@click.group(cls=AliasedGroup)
@click.version_option(package_name='dsFramework')
def cli():
    """
    DS framework cli

    ** How to use **

    g = generate

    Create project
    ==============
    dsf-cli g project my-new-project

    Generate forcer
    ===============
    dsf-cli g forcer my-new-forcer

    Generate predictable
    ====================
    dsf-cli g predictable my-new-predictable

    Create files
    ===================
    dsf-cli create-files [tester / deploy / cloud_eval / documentation / batch / trainer]

    Create Mlops dir
    ====================
    dsf-cli create-mlops

    run server
    ==========
    dsf-cli run-server

    run evaluation on a csv input
    =============================
    dsf-cli evaluate input-csv-file batch-size

    ==============
    Documentation
    ==============

    Generate documentation
    ======================
    dsf-cli generate documentation full

    Generate documentation - without base classes:
    ==============================================
    dsf-cli generate documentation simple

    Show documentation
    ==================
    dsf-cli show-documentation

    Upload documentation
    ====================
    dsf-cli upload-documentation

    ========================
    Batch project abilities
    ========================

    Workflow Creation
    =================
    dsf-cli generate workflow <new_workflow_name>

    DataProc Cluster Manipulation
    =============================
    dsf-cli dataproc-cluster <create / stop / start / delete / connect>

    Batch Stage Manipulation
    ========================
    dsf-cli batch-stage <create / run> my-stage

    Workflow Template Manipulation
    ==============================
    dsf-cli workflow-template <create / create_yaml / delete / instantiate>

    DAG Manipulation
    ================
    dsf-cli dag <create / import / delete>

    Upload files to bucket
    ======================
    dsf-cli upload-batch-files <jars, stages, whl , cluster_init, all>

    Miscellaneous
    Create sonar scanner properties file:
    ====================================
    dsf-cli create-sonar-file current_project_repo_name

    Install requirements for trainer
    ================================
    dsf-cli install-trainer-packages

    Install documentation doxygen
    =============================
    dsf-cli install-documentation-requirements

    Deploy project to GAE
    =====================
    dsf-cli deploy_service_to_gae
    """


# @click.option("--type", prompt="type", help="type of component")
# @click.option("--project_name", prompt="project_name", help="project_name")
# def apis(type, project_name):


@cli.command()
@click.argument('gen_type')
@click.argument('project_name')
def generate(gen_type, project_name):
    """! Build a generation function from the incoming gen_type.
        Args:
            gen_type: A string containing one of the following: [project / forcer / predictable / documentation / stage]
            project_name: A string declaring our project name
    """
    try:
        f = globals()["generate_%s" % gen_type]
    except Exception as e:
        click.echo('type ' + gen_type + ' not found')
        return
    f(project_name)


@cli.command()
@click.argument('file_type')
def create_files(file_type):
    """! Build a file creation function from the incoming file type.
        Args:
            file_type: A string containing one of the following:
                [tester / deploy / cloud_eval / documentation / batch / trainer]
    """
    try:
        f = globals()["create_%s_files" % file_type]
    except Exception as e:
        click.echo('type ' + file_type + ' not found')
        return
    f()


@cli.command()
def create_mlops():
    """! Copy Mlops directory from DSF """
    source_dir = os.path.join(__location__, 'dsframework/cli/mlops/')
    target_dir = os.path.join(os.path.abspath(os.getcwd()), 'mlops')
    shutil.copytree(source_dir, target_dir)


@cli.command()
def run_server():
    """! Start a uvicorn server by running /server/main.py """
    currentPipelineFolder = __proj_root__
    currentParentFolder = os.path.join(currentPipelineFolder, "..")
    os.environ["PYTHONPATH"] = currentPipelineFolder
    folder = currentPipelineFolder + '/server/main.py'
    subprocess.call('python ' + folder, shell=True)


@cli.command()
@click.argument('tag')
@click.argument('env', default="prd")
def cloud_eval(tag, env):
    # """Cloud evaluation running on AWS """
    currentPipelineFolder = __proj_root__
    currentParentFolder = os.path.join(currentPipelineFolder, "..")
    os.environ["PYTHONPATH"] = currentPipelineFolder
    folder = currentPipelineFolder + '/tester/dsp/get_model_options.py'
    output = subprocess.check_output('python ' + folder + ' ' + env, shell=True, encoding='utf-8').strip()
    if output:
        # print('output', output)
        l = output.split('\n')
        m = l[-1]
        try:
            l = ast.literal_eval(m)
            o = l[0]
            model = {
                'name': o['name'],
                'id': o['id'],
                'repo': o['repo'],
            }
            folder = currentPipelineFolder + '/tester/dsp/get_dataset_options.py'
            output = subprocess.check_output(
                'python ' + folder + ' ' + env + ' ' + model['name'] + ' ' + str(model['id']),
                shell=True, encoding='utf-8').strip()
            l = output.split('\n')
            m = l[-1]
            l = ast.literal_eval(m)
            selected = let_user_pick(l)
            # print('selected', selected)
            if not selected == None:
                selected_dataset = l[selected]
                # print('model',model)
                # print('selected_dataset', selected_dataset)
                folder = currentPipelineFolder + '/tester/dsp/add_experiment.py'
                subprocess.call(('dvc_push.sh'), shell=True)
                output = subprocess.call(('python', folder, env, json_dumps(model), json_dumps(selected_dataset), tag),
                                         shell=True)
                print('you can see your experiment in que')
            else:
                print('selection does not exist')
        except Exception as e:
            print(m)
    else:
        print('could not get models')


def create_deploy_files():
    """! Create deploy files (created through 'Create project', but can be used as a standalone function) """
    copy_deploy_files('')


def create_cloud_eval_files():
    """! Create cloud evaluation files (created through 'Create project', but can be used as a standalone function) """
    copy_cloud_eval_files(__proj_root__)


def create_tester_files():
    """! Create tester files (created through 'Create project', but can be used as a standalone function) """
    create_tester_env()


def create_documentation_files():
    """! Create documentation files (created through 'Create project', but can be used as a standalone function) """
    project_name = os.path.basename(os.getcwd())
    copy_documentation_files(os.path.join(directories['main'], 'documentation'))
    update_doc_configuration(project_name, os.path.join(directories['main'], 'documentation'))


def create_batch_files():
    """! Create batch files (created through 'Create project', but can be used as a standalone function) """
    copy_batch_files(directories['main'])


def create_trainer_files():
    """! Create trainer files (created through 'Create project', but can be used as a standalone function) """
    copy_trainer_files()


@cli.command()
@click.argument('wf_name')
@click.argument('action')
def dataproc_cluster(wf_name: str, action: str):
    """! Handle cluster action, either create,  start, stop, delete, or connect to the Dataproc cluster
            Args:
                wf_name: Relevant workflow that affects cluster configuration
                action: an action to perform, chosen from: {create, start, stop, delete, connect}
    """
    valid_actions = {'create', 'start', 'connect', 'stop', 'delete'}
    if action not in valid_actions:
        print(f"Can't determine action of type: {action}. Please use one of: {valid_actions}")
        return

    if not workflow_exists(wf_name):
        click.echo(f"Chosen workflow ({wf_name}) does not exist. Please choose a different workflow")
        return

    batch_helper = config_batch_helper(wf_name)

    if action == 'create':
        # Create DataProc cluster according to project specific configuration
        batch_helper.create_dataproc_cluster()
    elif action == 'start':
        # Start an existing DataProc cluster according to project specific configuration
        batch_helper.start_dataproc_cluster()
    elif action == 'connect':
        # Connect via ssh to an existing DataProc cluster according to project specific configuration.
        # Connection will be made to the master node
        batch_helper.connect_dataproc_cluster()
    elif action == 'stop':
        # Stop an existing DataProc cluster according to project specific configuration
        batch_helper.stop_dataproc_cluster()
    elif action == 'delete':
        # Delete an existing DataProc cluster according to project specific configuration.
        # Since we can't update certain parameters on an existing cluster, we need the option
        # to delete it
        batch_helper.delete_dataproc_cluster()


def update_config_file(stage_name: str, workflow_name: str, prerequisites: List[str]):
    """! Update relevant workflow configuration file with new stage
    @verbatim
    Args:
            stage_name: Stage name to be used
            workflow_name: Relevant workflow that will handle the new stage
            prerequisites: A list of stages that the new stage depends on
    @endverbatim
    """

    config_path = os.path.join(__proj_root__,
                               f'batch_files/workflows/{workflow_name}_workflow/config_{workflow_name}.json')

    stage_dict_entry = {"stage_name": stage_name,
                        "prerequisites": prerequisites}

    # Open json file, load to dictionary, update it and save it back
    with open(config_path, 'r+') as config_file:
        data = json_load(config_file)
        data['workflow_stages'].append(stage_dict_entry)
        config_file.seek(0)  # <--- should reset file position to the beginning.
        json_dump(data, config_file, indent=4)
        config_file.truncate()  # remove remaining part


def create_batch_stage(stage_name: str = '', workflow_name: str = ""):
    """! Generating new stage for a given workflow
    @verbatim
    Args:
            stage_name: Stage name to be used when creating files / folders / classes.
            workflow_name: Relevant workflow that will handle the new stage
    @endverbatim
    """
    stage_name = clean_name(stage_name)
    workflow_name = clean_name(workflow_name)

    if not validate_stage(stage_name, workflow_name):
        return

    create_stage_files_from_template(stage_name, workflow_name)

    # TBD: Should the prerequisites be validated, or this is a user responsibility?
    prerequisites = click.prompt('Please enter dependency stages (separate with comma)',
                                 type=str,
                                 default="",
                                 show_default=False)

    prerequisite_list = [] if prerequisites == "" else [list_value.strip() for list_value in prerequisites.split(',')]

    update_config_file(stage_name, workflow_name, prerequisite_list)

    click.echo(f'Created {stage_name} stage under {workflow_name} Workflow')


@cli.command()
@click.argument('action')
@click.argument('stage_name')
def batch_stage(action: str, stage_name: str):
    """! Handle batch stages action, either create or run a specific stage
        Args:
             action: an action to perform, chosen from: {create, create_yaml, instantiate, delete}
             stage_name: Name for the newly created stage, or the stage to be run on an existing cluster
    """
    valid_actions = {'create', 'run'}
    if action not in valid_actions:
        print(f"Can't determine action of type: {action}. Please use one of: {valid_actions}")
        return

    relevant_workflow = click.prompt('Please enter a valid workflow name', type=str)

    if action == 'create':
        # Create a batch stage
        create_batch_stage(stage_name, relevant_workflow)
    elif action == 'run':
        # Run specific stage on an existing Dataproc cluster
        batch_helper = config_batch_helper(relevant_workflow)
        batch_helper.run_single_stage(stage_name)


@cli.command()
@click.argument('wf_name')
@click.argument('action')
def workflow_template(wf_name: str, action: str):
    """! Handle workflow action, either create, instantiate or delete the template
        Args:
            wf_name: Relevant workflow to perform the selected action
            action: An action to perform, chosen from: {create, create_yaml, instantiate, delete}
    """
    valid_actions = {'create', 'create_yaml', 'delete', 'instantiate'}
    if action not in valid_actions:
        click.echo(f"Can't determine action type: {action}. Please use one of: {valid_actions}")
        return

    if not workflow_exists(wf_name):
        click.echo(f"Chosen workflow ({wf_name}) does not exist. Please choose a different workflow")
        return

    current_folder = __proj_root__

    os.environ["PYTHONPATH"] = current_folder
    workflow_folder = current_folder + '/batch_files/workflows/workflow.py'
    basic_command = 'python ' + workflow_folder

    click.echo(f"Attempting to {action} a workflow template")
    subprocess.call(basic_command + ' --wf-name ' + wf_name + ' --action ' + action, shell=True)


@cli.command()
@click.argument('wf_name')
@click.argument('action')
def dag(wf_name: str, action: str):
    """! Handle DAG actions, either create, delete, or import a DAG
    Once imported, it will start as paused and will be set up for a daily activation
        Args:
            wf_name: Relevant workflow to perform the selected action
            action: an action to perform, chosen from: {create, import, delete}
    """
    valid_actions = {'create', 'import', 'delete'}
    if action not in valid_actions:
        click.echo(f"Can't determine action type: {action}. Please use one of: {valid_actions}")
        return

    if not workflow_exists(wf_name):
        click.echo(f"Chosen workflow ({wf_name}) does not exist. Please choose a different workflow")
        return

    current_folder = __proj_root__

    os.environ["PYTHONPATH"] = current_folder
    dag_folder = current_folder + '/batch_files/workflows/dag.py'
    basic_command = 'python ' + dag_folder

    click.echo(f"Attempting to {action} a DAG")
    subprocess.call(basic_command + ' --wf-name ' + wf_name + ' --action ' + action, shell=True)


@cli.command()
@click.argument('wf_name')
@click.argument('file_type')
def upload_batch_files(wf_name: str, file_type: str):
    """! Upload all files needed for a batch project to the predefined bucket
        Args:
            wf_name: Relevant workflow to perform the selected action
            file_type: String containing the types of files to upload (jars, stages, whl , cluster_init, or "all")
    """
    if not workflow_exists(wf_name):
        click.echo(f"Chosen workflow ({wf_name}) does not exist. Please choose a different workflow")
        return

    batch_helper = config_batch_helper(wf_name)
    batch_helper.upload_files_to_bucket(file_type)


def config_batch_helper(wf_name: str):
    """! Configure the batch file helper for a variety of cli actions
        Args:
            wf_name: Relevant workflow to perform the selected action
        Returns:
            ZIDSBatchHelper: Instantiation of the batch helper class
    """
    from dsframework.cli.batch.batch_helper_functions import ZIDSBatchHelper

    batch_helper_cls = None

    with open(__proj_root__ + '/batch_files/batch_config.json') as batch_cfg_file:
        batch_config = json_load(batch_cfg_file)

    with open(__proj_root__ + f"/batch_files/workflows/{wf_name}_workflow/config_{wf_name}.json") as wf_cfg_file:
        wf_config = json_load(wf_cfg_file)

    config_dict = {**batch_config, **wf_config}

    batch_helper_cls = ZIDSBatchHelper(config_dict)

    return batch_helper_cls


@cli.command()
@click.argument('repo_name')
def create_sonar_file(repo_name: str):
    """! Create a sonar scanner file.

    Args:
        repo_name: name of your repository.
    """

    data = "sonar.projectKey=dozi_" + repo_name
    create_file(os.path.join(__proj_root__, "sonar-scanner.properties"), data)


@cli.command()
@click.argument('csv_file_path')
@click.argument('batch_size')
def evaluate(csv_file_path, batch_size):
    """! Quick 'Initial test' to check the module.

    @verbatim
    Args:
        csv_file_path: CSV file path.
        batch_size: The size that will be taken every iteration from the CSV - in order to feed the model.
    @endverbatim
    """

    if os.path.isfile(csv_file_path):
        current_pipeline_folder = __proj_root__
        os.environ["PYTHONPATH"] = current_pipeline_folder
        folder = current_pipeline_folder + '/tester/general_tester.py'
        subprocess.call('python ' + folder + ' ' + csv_file_path + ' ' + batch_size, shell=True)
    else:
        click.echo('file: ' + csv_file_path + ' not found')


@cli.command()
def show_documentation():
    """! Opens a browser and show the local project documentation."""
    launch_doc_webbrowser()


def launch_doc_webbrowser():
    current_pipeline_folder = __proj_root__

    current_doc_folder = os.path.join(current_pipeline_folder, "documentation")
    if os.path.isdir(os.path.join(current_pipeline_folder, "dsframework", "documentation")):
        current_doc_folder = os.path.join(current_pipeline_folder, "dsframework", "documentation")

    project_html_path = os.path.join(current_doc_folder, 'project_docs', 'html', 'index.html')
    if os.path.isfile(project_html_path):
        webbrowser.open('file://' + project_html_path)
    else:
        click.echo('Project documentation not found, please generate documentation using: \'dsf-cli generate '
                   'documentation full\'')


ALIASES = {
    "g": generate
}


def generate_project(project_name):
    """! Generating new project (with the relevant project_name)\n
    How to generate project :
    > dsf-cli g project my-new-project
    @verbatim
    Args:
            project_name: Project name will be used in the creation of the files.
    @endverbatim
    """
    # project_name = clean_name(project_name)
    click.echo('Generating project: ' + project_name)
    create_project(project_name)


def generate_forcer(file_name):
    """! Generate new forcer - can help us, if we need forcer on an existing project """
    file_name = clean_name(file_name)
    create_exist_pipeline_file('forcer', file_name)


def generate_predictable(file_name):
    """! Generate new predictable - can help us, if we need predictable on an existing project
    """
    file_name = clean_name(file_name)
    create_exist_pipeline_file('predictable', file_name)


def add_workflow_to_setup(workflow_name: str):
    """! Modify setup.py for the whl creation of batch projects
    @verbatim
    Args:
            workflow_name: Workflow to be added to the file
    @endverbatim
    """

    insert_string = f"if build_type == '{workflow_name}':\n\tdata = config + pipeline + batch_files + tester\n\n"
    setup_file_path = os.path.join(__proj_root__, "batch_files/setup.py")

    with open(setup_file_path, 'r+') as setup_fd:
        contents = setup_fd.readlines()
        if "# Additional workflows" in contents[-1]:  # Handle last line to prevent IndexError
            contents.append(insert_string)
        else:
            for index, line in enumerate(contents):
                if "# Additional workflows" in line and insert_string not in contents[index + 1]:
                    contents.insert(index + 1, insert_string)
                    break
        setup_fd.seek(0)
        setup_fd.writelines(contents)


def generate_workflow(workflow_name: str):
    """! Generating a new workflow to be used on GCP Dataproc
    @verbatim
    Args:
            workflow_name: Workflow to be created
    @endverbatim
    """

    click.echo('Generating workflow: ' + workflow_name)

    if not os.path.exists(os.path.join(__proj_root__, "batch_files")):
        click.echo("Couldn't find batch project folder, creating...")
        copy_batch_files(directories['main'])

    if workflow_exists(workflow_name):
        click.echo(f'Workflow "{workflow_name}" already exists. Please use a different name.')
        return

    workflow_path = os.path.join(__proj_root__, f"batch_files/workflows/{workflow_name}_workflow")

    # Create primary workflow folder
    os.mkdir(workflow_path)
    create_empty_init_file(workflow_path)

    # Create an empty stages folder
    stages_path = os.path.join(workflow_path, "stages")
    os.mkdir(stages_path)
    create_empty_init_file(stages_path)

    # Create an empty requirement file
    create_file(os.path.join(workflow_path, f"requirements_{workflow_name}.txt"), "")

    # Create configuration file from template
    data = read_template_file('batch/workflow_config', "json")
    data = data.replace('{name-your-wf}', workflow_name)
    data = data.replace('{name-your-service}', os.path.split(os.getcwd())[-1])

    create_file(os.path.join(workflow_path, f"config_{workflow_name}.json"), data)
    
    # Manipulate setup.py to include the new build type
    add_workflow_to_setup(workflow_name)


def workflow_exists(wf_name: str = ""):
    """! Checks if a workflow with the given name exists
        Args:
                wf_name: Workflow name to be validated
        Returns:
            True if workflow already exists, False otherwise
        """

    return os.path.exists(os.path.join(__proj_root__, f'batch_files/workflows/{wf_name}_workflow'))


def validate_stage(stage_name: str = "", workflow: str = ""):
    """! Checks if a stage with the given name is valid
    Args:
            stage_name: Stage name to be validated
            workflow: Workflow name to be validated
    Returns:
        True if stage can be created correctly, False otherwise
    """

    # Check workflow validity
    if not workflow_exists(workflow):
        click.echo(f'Workflow "{workflow}" does not exist. Please create or try a different workflow.')
        return False

    # Check stage validity
    if os.path.exists(os.path.join(__proj_root__, f'batch_files/workflows/{workflow}_workflow/stages/{stage_name}')):
        click.echo(f'Stage "{stage_name}" already exists for given workflow "{workflow}". Please use a different name.')
        return False

    return True


def generate_documentation(simple="full"):
    """! Document project CLI

        Usage:
            dsf-cli generate documentation full
    """
    if simple == 'full':
        click.echo('Documenting project .....')
    else:
        click.echo('Documenting project (without base).....')

    # Get folders
    current_pipeline_folder = __proj_root__
    current_doc_folder = os.path.join(current_pipeline_folder, "documentation")
    if os.path.isdir(os.path.join(current_pipeline_folder, "dsframework", "documentation")):
        current_doc_folder = os.path.join(current_pipeline_folder, "dsframework", "documentation")

    # Validity checks
    if not os.path.isfile(os.path.join(current_doc_folder, 'project_doxyfile')):
        click.echo("Documentation folder not found, please change directory to project root folder.")
        return

    venv_path = os.getenv('VIRTUAL_ENV')

    try:
        if simple == 'full':
            os.environ["EXCLUDE_PATH"] = get_exclude_string(venv_path)
        else:
            os.environ["EXCLUDE_PATH"] = venv_path
    except Exception as e:
        click.echo(f'Switching to simple: {e}')
        os.environ["EXCLUDE_PATH"] = venv_path

    click.echo(f'Creating project documentation in {os.path.join(current_doc_folder, "project_docs")}')

    try:
        # Run doxygen as subprocess
        subprocess.call('doxygen project_doxyfile', shell=True, cwd=current_doc_folder)
    except Exception as e:
        click.echo(f'Project documentation failed!, exception: {e}')
    else:
        click.echo('Project documentation done!')
        launch_doc_webbrowser()


def get_exclude_string(venv_path):
    packages_path = site.getsitepackages()  # Get packages folder.

    # Get list of packages to exclude, except 'dsframework'
    pkg_exclude_folders = get_excluded_path_list('(?!.*dsframework)', packages_path[0], 'folders')

    # Get list of scattered file in packages folder
    pkg_exclude_files = get_excluded_path_list('(?!.*api_cli.py)', packages_path[0], 'files')

    # Get a list of files and folders in dsframework, except 'base' folder.
    pkg_exclude_dsf = get_excluded_path_list('(?!.*base)', os.path.join(packages_path[0], 'dsframework'), 'both')

    # Get venv all the rest the folders to exclude, except 'venv/lib'
    exclude_venv = get_excluded_path_list('(?!.*lib)', venv_path, 'both')

    # Combine two lists to one string: exclude_folders and exclude_folders
    exclude_full_list = pkg_exclude_folders + exclude_venv + pkg_exclude_files + pkg_exclude_dsf
    exclude_string = " \\ ".join(map(str, exclude_full_list))  # Seperated by backslash

    return exclude_string


def get_excluded_path_list(regex_str, path, get_type):
    if get_type == 'folders':
        path_list = [os.path.join(path, name) for name in os.listdir(path)
                     if os.path.isdir(os.path.join(path, name))]
    elif get_type == 'files':
        path_list = [os.path.join(path, name) for name in os.listdir(path)
                     if os.path.isfile(os.path.join(path, name))]
    else:  # 'both'
        path_list = [os.path.join(path, name) for name in os.listdir(path)]

    if regex_str != '':
        regex = re.compile(regex_str)
        return list(filter(regex.match, path_list))
    else:
        return path_list


def zip_folder(local_documentation_folder, path_file_name):

    if os.path.isdir(local_documentation_folder):
        if os.listdir(local_documentation_folder):
            shutil.make_archive(path_file_name, 'zip', local_documentation_folder)


def hash_file(file_path):

    hash_code = hashlib.sha256()

    with open(file_path, 'rb') as f:
        while True:
            chunk = f.read(4096)
            if not chunk:
                break
            hash_code.update(chunk)

    return hash_code.hexdigest()


@cli.command()
def upload_documentation():
    """! Upload project documentation CLI

        Usage:
            dsf-cli upload-documentation
    """

    # Set folders
    current_working_path = os.getcwd()
    project_name = os.path.split(current_working_path)[1]
    current_doc_folder = os.path.join(current_working_path, 'documentation', 'project_docs')
    local_upload_folder = os.path.join(current_working_path, 'documentation', 'upload')
    html_folder_name = 'html'

    google_gs_path = 'gs://dozi-stg-ds-apps-1-ds-apps-ds-portal/documentation/'

    project_gs_folder = f'{google_gs_path}'

    if project_name == 'ds-framework':
        current_doc_folder = os.path.join(current_working_path, 'dsframework', 'documentation', 'project_docs')
        local_upload_folder = os.path.join(current_working_path, 'dsframework', 'documentation', 'upload')

    # Validity checks
    if not os.path.isfile(os.path.join(current_doc_folder, html_folder_name, 'index.html')):
        click.echo('Documentation folder not found, please change directory to project root folder.')
        return

    local_documentation_folder = os.path.join(current_doc_folder, html_folder_name)

    if os.path.isdir(local_upload_folder):
        shutil.rmtree(local_upload_folder)

    os.mkdir(local_upload_folder)

    zip_folder(local_documentation_folder, os.path.join(local_upload_folder, project_name))

    hash_code = hash_file(os.path.join(local_upload_folder, project_name) + '.zip')
    with open(os.path.join(local_upload_folder, f'{project_name}_hash.txt'), 'w') as file_hash:
        file_hash.write(hash_code)

    # Upload current documentation
    click.echo(f'Uploading {project_name} documentation started....')

    call_gsutil_command(f'gsutil -m cp -r {local_upload_folder} {project_gs_folder}',
                        project_name, current_doc_folder, 'Uploading documentation')

    click.echo(f'Uploading {project_name} documentation done!!!')


def call_gsutil_command(gsutil_command, project_name, working_folder, command_desc):
    click.echo(f'{command_desc} {project_name} ....')

    try:
        subprocess.call(gsutil_command, shell=True, cwd=working_folder)
    except Exception as e:
        click.echo(f'{command_desc} failed with exception: {e}')
        return
    else:
        click.echo(f'{command_desc} done!')


def clean_name(name):
    """! Change the name from 'word1-word2' --> 'word1_word2' """
    name = name.replace('-', '_')
    return name


def create_folders(project_name):
    """! Create all needed folders for the new project creation
    @verbatim
    Args:
        project_name: Project name
    @endverbatim
    """

    global directories
    directories['main'] = project_name
    if not os.path.exists(directories['main']):
        os.mkdir(directories['main'])

    create_main_folders('config', 'main')
    create_main_folders('pipeline', 'main')
    create_main_folders('artifacts', 'pipeline')
    create_main_folders('models', 'artifacts')
    create_main_folders('vocabs', 'artifacts')
    # create_main_folders('other', 'artifacts')
    create_main_folders('preprocessor', 'pipeline')
    create_main_folders('predictables', 'pipeline')
    create_main_folders('predictors', 'pipeline')
    create_main_folders('forcers', 'pipeline')
    create_main_folders('postprocessor', 'pipeline')
    create_main_folders('schema', 'pipeline')
    create_main_folders('tester', 'main')
    create_main_folders('test_schema', 'tester')
    create_main_folders('dsp', 'tester')
    create_main_folders('trainer', 'main')
    create_main_folders('training_output', 'trainer')
    create_main_folders('pl_wrapper', 'trainer')
    create_main_folders('server', 'main')


def create_project(project_name):
    """! Create project from templates.
    @verbatim
    Args:
        project_name: Project name
    @endverbatim
    """

    create_folders(project_name)
    original_project_name = project_name
    project_name = clean_name(project_name)

    # Create pipeline files
    create_pipeline_file(project_name, directories['artifacts'], 'shared_artifacts')
    create_pipeline_file(project_name, directories['preprocessor'], 'preprocess')
    create_pipeline_file(project_name, directories['predictors'], 'predictor')
    create_pipeline_file(project_name, directories['forcers'], 'forcer')
    create_pipeline_file(project_name, directories['postprocessor'], 'postprocess')
    create_pipeline_file(project_name, directories['predictables'], 'predictable')
    create_pipeline_file(project_name, directories['pipeline'], 'pipeline')
    create_pipeline_file(project_name, directories['main'], 'pipeline_test', False)

    # Create schema files
    create_schema_file(project_name, directories['schema'], 'inputs')
    create_schema_file(project_name, directories['schema'], 'outputs')
    create_schema_file(project_name, directories['schema'], '__init__', False)

    # Create tester files
    create_tester_file(project_name, directories['tester'], 'general_tester', False)
    create_tester_file(project_name, directories['tester'], 'evaluator', False)
    create_empty_init_file(directories['tester'])
    create_schema_file(project_name, directories['test_schema'], 'test_input', False)
    create_schema_file(project_name, directories['test_schema'], 'test_output', False)
    create_empty_init_file(directories['test_schema'])
    create_dsp_file(original_project_name, directories['dsp'], 'get_model_options', False)
    create_dsp_file(original_project_name, directories['dsp'], 'get_dataset_options', False)
    create_dsp_file(original_project_name, directories['dsp'], 'add_experiment', False)
    create_empty_init_file(directories['dsp'])

    # Create trainer files
    create_trainer_file(project_name, directories['pl_wrapper'], 'custom_dataset', False, from_folder='pl_wrapper')
    create_trainer_file(project_name, directories['pl_wrapper'], 'data_module', False, from_folder='pl_wrapper')
    create_trainer_file(project_name, directories['pl_wrapper'], 'iterable_dataset', False, from_folder='pl_wrapper')
    create_trainer_file(project_name, directories['pl_wrapper'], 'plmodel', False, from_folder='pl_wrapper')
    create_trainer_file(project_name, directories['pl_wrapper'], 'network_module', False, from_folder='pl_wrapper')
    create_trainer_file(project_name, directories['pl_wrapper'], 'save_onnx', False, from_folder='pl_wrapper')

    create_empty_init_file(directories['pl_wrapper'])
    create_trainer_file(project_name, directories['trainer'], 'comparison', False)
    create_trainer_file(project_name, directories['trainer'], 'config', False)
    create_trainer_file(project_name, directories['trainer'], 'data', False)
    create_trainer_file(project_name, directories['trainer'], 'dataset_preparation', False)
    create_trainer_file(project_name, directories['trainer'], 'model', False)
    create_trainer_file(project_name, directories['trainer'], 'test', False)
    create_trainer_file(project_name, directories['trainer'], 'train', False)
    create_trainer_file(project_name, directories['trainer'], 'main', False)
    create_empty_init_file(directories['trainer'])
    source_path = os.path.join(__location__, 'dsframework/cli/trainer/requirements_trainer.txt')
    shutil.copyfile(source_path, directories['trainer'] + '/requirements_trainer.txt')

    # Create server files
    create_server_file(project_name, directories['server'], 'main', False)
    create_server_file(project_name, directories['server'], 'test_server_post', False)
    create_server_file(project_name, directories['server'], 'pool', False)
    create_server_file(project_name, directories['server'], 'token_generator', False)
    create_server_file(project_name, directories['server'], '__init__', False)

    create_project_config_json()
    create_project_gitignore()
    create_server_config_json()
    copy_deploy_files(directories['main'])
    copy_cloud_eval_files(directories['main'])
    copy_batch_files(directories['main'])
    copy_documentation_files(os.path.join(directories['main'], 'documentation'))
    update_doc_configuration(project_name, os.path.join(directories['main'], 'documentation'))

    change_to_project_dir()
    run_dvc_init()
    run_git_init()


def create_main_folders(target_dir, base_dir):
    """! Create folder if not already exists
    @verbatim
    Args:
        target_dir: the new folder to create
        base_dir  : where we want to create the new folder.
    @endverbatim
    """

    global directories
    directories[target_dir] = directories[base_dir] + '/' + target_dir
    if not os.path.exists(directories[target_dir]):
        os.mkdir(directories[target_dir])


def create_pipeline_file(project_name, folder, pipeline_type, create_init=True):
    """! Takes the template file and replace it with new one with project_name
    @verbatim
    Args:
        project_name : The project name
        folder       : The folder we want to create the new file.
        pipeline_type : File type in the pipeline.
        create_init(=True): Create init file - if doesnt exist.
    @endverbatim
    """

    data = read_template_file(pipeline_type)
    replace_in_template_and_create_file(project_name, folder, pipeline_type, data, create_init)


def create_schema_file(project_name, folder, pipeline_type, create_init=True):
    """! Create schema files (read the template and replace it )
    @verbatim
    Args:
        project_name : The project name
        folder       : The folder we want to create the new file.
        pipeline_type : File type. TODO change pipelineType name
        create_init(=True): Create init file - if doesnt exist.
    @endverbatim
    """
    data = read_template_file('schema/' + pipeline_type)
    replace_in_template_and_create_file(project_name, folder, pipeline_type, data, create_init)


def create_server_file(project_name, folder, pipeline_type, create_init=True):
    """! Create server file (read the template and use him to create the new file)
    @verbatim
    Args:
        project_name : The project name
        folder       : The folder we want to create the new file.
        pipeline_type : File type. TODO change pipelineType name
        create_init(=True): Create init file - if doesnt exist.
    @endverbatim
    """
    data = read_template_file('tester/server/' + pipeline_type)
    replace_in_template_and_create_file(project_name, folder, pipeline_type, data, create_init)


def create_tester_file(project_name, folder, test_file_type, create_init=True):
    """! Create tester file (read the template and use him to create the new file)
    @verbatim
    Args:
        project_name   : The project name.
        folder         : The folder we want to create the new file.
        test_file_type : File type.
        create_init(=True): Create init file - if doesnt exist.
    @endverbatim
    """
    data = read_template_file('tester/' + test_file_type)
    replace_in_template_and_create_file(project_name, folder, test_file_type, data, create_init)


def create_trainer_file(project_name, to_folder, test_file_type, create_init=True, from_folder=''):
    """! Create trainer file (read the template and use him to create the new file)
    @verbatim
    Args:
        project_name   : The project name.
        to_folder      : The folder we want to create the new file in.
        test_file_type : File type.
        create_init(=True): Create init file - if it doesn't exist.
        from_folder    : The folder from which we read the template file
    @endverbatim
    """

    # data = read_template_file('trainer/' + test_file_type)
    data = read_template_file(os.path.join('trainer', from_folder, test_file_type))
    replace_in_template_and_create_file(project_name, to_folder, test_file_type, data, create_init)


def create_dsp_file(project_name, folder, test_file_type, create_init=True):
    """! Create DSP file (read the template and use him to create the new file)
    @verbatim
        Args:
            project_name   : The project name.
            folder         : The folder we want to create the new file.
            test_file_type : File type.
            create_init(=True): Create init file - if doesnt exist.
    @endverbatim
    """
    data = read_template_file('tester/dsp/' + test_file_type)
    replace_in_template_and_create_file(project_name, folder, test_file_type, data, create_init)


def create_exist_pipeline_file(type, fileName):
    """! Create a new component in the pipeline and integrate it (for example forcer or predictable)
    @verbatim
    Args:
        type      : New component type (forcer or predictable)
        fileName  : New file name.
    @endverbatim
    """

    pipelineType = type + 's'
    folder = 'pipeline/' + pipelineType
    if os.path.exists(folder):
        data = read_template_file(type)
        fileNameNoUnderscore = to_capitalize_no_underscore(fileName)
        className = fileNameNoUnderscore + type.capitalize()
        currentPipelineFolder = os.path.basename(os.getcwd())
        currentDir = folder.replace('/', '.')

        data = data.replace('generatedClass', className)

        new_file = folder + "/" + fileName + ".py"
        current_init_file = folder + "/__init__.py"
        new_init_export = "from " + '.' + fileName + " import " + className

        create_file(new_file, data)
        create_init_file(current_init_file, new_init_export)
        inject_to_pipeline(fileName, type, className, new_init_export)
    else:
        print('please create a project and go to project location first')
    pass


def create_stage_files_from_template(stage_name: str = "", workflow_name: str = ""):
    """! Create all needed folders and files for the new stage
    @verbatim
    Args:
        stage_name: Stage name
        workflow_name: Workflow that this stage belongs to
    @endverbatim
    """

    data = read_template_file('batch/stage')
    stage_directory_name = os.path.join(__proj_root__,
                                        f'batch_files/workflows/{workflow_name}_workflow/stages/{stage_name}')
    os.mkdir(stage_directory_name)
    replace_in_stage_template_and_create_file(stage_name, stage_directory_name, workflow_name, data)


def create_tester_env():
    """! This function will create a testing environment, if one does not already exist in the project.
    It will create the testing, evaluation, and relevant schema files.
    """

    project_name = os.getcwd().split('/')[-1]
    dst_folder = 'tester/'
    schema_dst_folder = 'tester/test_schema/'
    dsp_dst_folder = 'tester/dsp/'
    tester_file_list = ["tester/general_tester",
                        "tester/evaluator",
                        "schema/test_input",
                        "schema/test_output",
                        "tester/dsp/add_experiment",
                        "tester/dsp/get_dataset_options",
                        "tester/dsp/get_model_options"]

    if not os.path.exists(dst_folder):
        os.mkdir(dst_folder)
        os.mkdir(schema_dst_folder)
        os.mkdir(dsp_dst_folder)

        for filename_with_path in tester_file_list:
            file_only = filename_with_path.split('/')[-1]
            parent_folder = filename_with_path.split('/')[-2]
            if parent_folder == 'schema':
                dst_folder = schema_dst_folder
            elif parent_folder == 'dsp':
                dst_folder = dsp_dst_folder

            data = read_template_file(filename_with_path)

            clean_project_name = clean_existing_project_name(project_name)
            filename_no_underscore = to_capitalize_no_underscore(file_only)

            class_name = clean_project_name + filename_no_underscore

            data = data.replace('generatedClass', class_name)
            data = data.replace('generatedProjectName', clean_project_name)

            new_file = dst_folder + "/" + file_only + ".py"
            create_file(new_file, data)

            # Test schema does not use generated class name for now, so this code should be modified before uncommenting
            # current_init_file = dst_folder + "/" + "__init__.py"
            # new_init_export = "from " + dst_folder.replace('/', '.') + "." + file_only + " import " + class_name
            # create_init_file(current_init_file, new_init_export)
    else:
        print('Tester folder Already exists. If you wish to recreate, please erase folder and retry')
    pass


def read_template_file(filename: str, file_type: str = "py"):
    """! Read \<filename\>_template.py  from the cli dir """
    with open(os.path.join(__location__, 'dsframework/cli/' + filename + f"_template.{file_type}"), 'r') as file:
        data = file.read()
        return data


def read_file(file_path):
    with open(file_path, 'r') as file:
        data = file.read()
        return data


def write_to_file(file_path, data):
    with open(file_path, 'w') as file:
        file.write(data)
        file.close()


def replace_in_template_and_create_file(project_name, folder, file_name, data, with_init_file):
    """! Get & Replace the template file and create new one with project_name
    @verbatim
    Args:
        project_name : The project name
        folder       : The folder we want to create the new file.
        file_name : Which File.
        data         : The template file.
        with_init_file(=True): Create init file if doesn't exist.
    @endverbatim
    """

    pipeline_type_no_underscore = to_capitalize_no_underscore(file_name)
    project_name_no_underscore = to_capitalize_no_underscore(project_name)
    class_name = project_name_no_underscore + pipeline_type_no_underscore
    class_name_for_base_object = project_name_no_underscore + pipeline_type_no_underscore
    current_dir = folder.replace('/', '.')
    current_base_dir = directories['main'].replace('/', '.')
    current_pipeline_dir = directories['pipeline'].replace('/', '.')

    data = data.replace('generatedClassName', class_name_for_base_object)
    data = data.replace('generatedClass', class_name)
    data = data.replace('original_project_name', project_name)
    data = data.replace('generatedProjectName', project_name_no_underscore)
    data = data.replace('generatedDirectory', current_dir)
    data = data.replace('generatedBaseDir', current_base_dir)
    data = data.replace('generatedPipelineDir', current_pipeline_dir)

    new_file = folder + "/" + file_name + ".py"
    create_file(new_file, data)

    if with_init_file:
        new_init_file = folder + "/__init__.py"
        new_init_export = "from " + '.' + file_name + " import " + class_name
        create_init_file(new_init_file, new_init_export)


def replace_in_stage_template_and_create_file(stage_name: str, folder: str, workflow_name: str, data):
    """! Get & Replace the stage template file and create a new one with stage_name
    @verbatim
    Args:
        stage_name : The stage name
        folder     : The destination folder for the new files
        workflow_name : Workflow name under which the stage will run
        data       : The template file
    @endverbatim
    """

    data = data.replace('generatedStageName', to_capitalize_no_underscore(stage_name))
    data = data.replace('generatedWFName', workflow_name)

    create_file(folder + "/main.py", data)
    create_empty_init_file(folder)


def create_empty_init_file(folder: str):
    """Creates an empty __init__.py file
    Args:
        folder       : The folder we want to create the new file in.
    """

    new_init_file = folder + "/__init__.py"
    create_init_file(new_init_file, '')


def create_predictable_file(fileName):
    pass


def create_file(new_file_path, data):
    """! Create new file and write the data file into it
    @verbatim
    Args:
        new_file_path : New file name path.
        data          : The content of the new file.
    @endverbatim
    """

    if not os.path.exists(new_file_path):
        f = open(new_file_path, "w")
        f.write(data)
        f.close()


def create_init_file(init_path, init_export):
    """! if init file doesn't exist - create it and add the export line
         if init file exist - check if the init_export line appears in the init file : if not add it.

    @verbatim
    Args:
        init_path   : Init file.
        init_export : Import line to add if needed.
    @endverbatim
    """

    if not os.path.exists(init_path):
        f = open(init_path, "w")
        f.write(init_export)
        f.close()
    else:
        f = open(init_path, 'r+')
        data = f.read()
        if init_export not in data:
            if len(data) and not data.endswith('\n'):
                f.write('\n')
            f.write(init_export)
            f.close()


def save_file(file_name_path, data):
    f = open(file_name_path, "w")
    f.write(data)
    f.close()


def inject_to_pipeline(fileName, type, className, new_init_export):
    """! Building the component in the pipeline file
    @verbatim
    Args:
        fileName       : New file name.
        type           : New component type (forcer or predictable)
        className      : New component class.
        new_init_export: Create init file - if doesnt exist.
    @endverbatim
    """

    file_path = 'pipeline/pipeline.py'
    if os.path.exists(file_path):
        data = read_file(file_path)
        new_tab_line = '\n'
        data_changed = False
        last_index_of_import = -1
        first_index_of_add_component = -1
        index_of_class = re.search(r'class[^\n]*', data)

        # finding current indent config
        index_of_build_pipeline = re.search(r'def build_pipeline[^\n]*', data)
        index_of_build_preprocessor = re.search(r'self.preprocessor =[^\n]*', data)
        index_of_build_postprocessor = re.search(r'self.postprocessor =[^\n]*', data)
        index_of_build_pipeline_row = re.search(r'.*def build_pipeline[^\n]*', data)
        if index_of_build_pipeline:
            build_pipeline_indent = (index_of_build_pipeline.start() - index_of_build_pipeline_row.start()) * 2
            new_tab_line = new_tab_line.ljust(build_pipeline_indent + 1)

        attr = fileName + type.title()
        new_component_line = 'self.' + attr + ' = ' + className + '()'
        add_component_line = 'self.add_component(self.' + attr + ')'

        # finding imports and add components indexes
        last_index_of_add_component = -1
        all_from_import = [i.end() for i in re.finditer(r'from[^\n]*', data)]
        all_add_components = [[i.start(), i.end()] for i in re.finditer(r'self.add_component[^\n]*', data)]
        if len(all_from_import):
            last_index_of_import = all_from_import[-1]
        if len(all_add_components):
            first_index_of_add_component = all_add_components[0][0]
            last_index_of_add_component = all_add_components[-1][-1]
        # finding imports and add components indexes

        index_to_add = 0

        # add import to end of imports or to top of file
        if last_index_of_import > -1 and new_init_export not in data:
            s = '\n' + new_init_export
            index_to_add += len(s)
            data = data[:last_index_of_import] + s + data[last_index_of_import:]
            data_changed = True
        elif index_of_class and new_init_export not in data:
            s = new_init_export + '\n\n'
            index_to_add += len(s)
            data = data[:index_of_class.start()] + s + data[index_of_class.start():]
            data_changed = True

        # check if build_pipeline exist but with no components yet
        if first_index_of_add_component == -1 and last_index_of_add_component == -1 and index_of_build_pipeline:
            s = new_tab_line
            current_end = index_of_build_pipeline.end()
            if index_of_build_preprocessor:
                current_end = index_of_build_preprocessor.end()
            if index_of_build_postprocessor:
                current_end = index_of_build_postprocessor.end()
            index = current_end + index_to_add
            index_to_add += len(s)
            data = data[:index] + s + data[index:]
            data_changed = True
            first_index_of_add_component = current_end
            last_index_of_add_component = current_end

        # adding new component line
        if first_index_of_add_component > -1 and new_component_line not in data:
            first_index_of_add_component += index_to_add
            if len(all_add_components):
                new_component_line = new_component_line + new_tab_line
            index_to_add += len(new_component_line)
            data = data[:first_index_of_add_component] + new_component_line + data[first_index_of_add_component:]
            data_changed = True

        # adding add_component line
        if last_index_of_add_component > -1 and add_component_line not in data:
            last_index_of_add_component += index_to_add
            add_component_line = new_tab_line + add_component_line
            index_to_add += len(add_component_line)
            data = data[:last_index_of_add_component] + add_component_line + data[last_index_of_add_component:]
            data_changed = True

        if data_changed:
            write_to_file(file_path, data)


def create_project_config_yaml():
    """! Creates /pipeline/config.yaml from dsframework/cli/config.yaml """
    with open(os.path.join(__location__, 'dsframework/cli/config.yaml'), 'r') as file:
        data = file.read()
        data = data.replace('generatedDirectory', directories['main'])
        new_file = directories['main'] + '/pipeline/config.yaml'
        if not os.path.exists(new_file):
            f = open(new_file, "w")
            f.write(data)
            f.close()


def create_project_config_json():
    """! Creates /config/config.json from dsframework/cli/config.json """
    with open(os.path.join(__location__, 'dsframework/cli/config.json'), 'r') as file:
        data = file.read()
        data = data.replace('generatedDirectory', directories['main'])
        new_file = directories['main'] + '/config/config.json'
        if not os.path.exists(new_file):
            f = open(new_file, "w")
            f.write(data)
            f.close()


def create_project_gitignore():
    """! Creates /.gitignore from dsframework/cli/.gitignore
         Which means all files that appear in the project directory and should not be uploaded to the git repository.
    """
    with open(os.path.join(__location__, 'dsframework/cli/.gitignore'), 'r') as file:
        data = file.read()
        new_file = directories['main'] + '/.gitignore'
        if not os.path.exists(new_file):
            f = open(new_file, "w")
            f.write(data)
            f.close()


def create_server_config_json():
    """! Creates server/cors_allowed_origins from dsframework/cli/cors_allowed_origins   """
    with open(os.path.join(__location__, 'dsframework/cli/cors_allowed_origins.json'), 'r') as file:
        data = file.read()
        data = data.replace('generatedDirectory', directories['main'])
        new_file = directories['server'] + '/cors_allowed_origins.json'
        if not os.path.exists(new_file):
            f = open(new_file, "w")
            f.write(data)
            f.close()


def change_to_project_dir():
    """! Change the OS dir path """
    os.chdir(directories['main'])


def run_dvc_init():
    """! Running dvc init and doing the following:
    1. Defines where all the dvc will be uploaded (DVC_BUCKET=)
    2. dvc init    - generate a local repository
    3. dvc config
    4. dvc remote add - connect my local repo -> remote
    5. dvc remote modify
    """
    dir_path = os.getcwd()
    if not os.path.isdir(dir_path + '/.dvc'):
        command = 'dvc_init.sh'
        if not isWindows:
            command = './' + command
        subprocess.call(command, shell=True)


def run_git_init():
    """! Using git_init.sh to create a new project on git
        https://git.zoominfo.com/dozi/{name-your-service}
    """
    dir_path = os.getcwd()
    if not os.path.isdir(dir_path + '/.git'):
        command = 'git_init.sh'
        if not isWindows:
            command = './' + command
        subprocess.call(command, shell=True)


def copy_deploy_files(main_dir):
    """! Copy deploy files (from dsframework/cli/tester/deploy_files """
    currentPipelineFolder = os.path.basename(os.getcwd())
    if main_dir:
        currentPipelineFolder = main_dir
    dir = os.path.join(__location__, 'dsframework/cli/tester/deploy_files/')
    listOfFiles = list()
    for (dirpath, dirname, filenames) in os.walk(dir):
        dirpath = dirpath.replace(dir, '')
        listOfFiles = [os.path.join(dirpath, file) for file in filenames]
        for file_path in listOfFiles:
            with open(os.path.join(dir, file_path), 'r', encoding="utf8") as file:
                try:
                    data = file.read()
                    data = data.replace('{name-your-service}', currentPipelineFolder)
                    data = data.replace('{name-your-artifacts}', currentPipelineFolder)
                    dirToCreate = ''
                    if main_dir:
                        dirToCreate = os.path.join(main_dir, dirpath)
                    if not os.path.exists(dirToCreate):
                        os.makedirs(dirToCreate)
                    if not os.path.exists(os.path.join(main_dir, file_path)):
                        f = open(os.path.join(main_dir, file_path), "w")
                        f.write(data)
                        f.close()
                        # Make sure that .sh files can be executed
                        if (os.path.join(main_dir, file_path)).endswith('.sh'):
                            os.chmod(dirToCreate + file_path, 0o744)
                except Exception as ex:
                    pass


def copy_cloud_eval_files(main_dir):
    """! Copy cloud evaluation files """
    currentPipelineFolder = os.path.basename(os.getcwd())
    if main_dir:
        currentPipelineFolder = main_dir
    service_folder = str(currentPipelineFolder)
    service_name = service_folder.split('/')[-1]
    project_name_no_underscore = clean_existing_project_name(service_name)
    cloud_eval_dir = os.path.join(__location__, 'dsframework/cli/tester/cloud_eval/')
    listOfFiles = list()
    conflicted_files = []
    for (dirpath, dirname, filenames) in os.walk(cloud_eval_dir):
        if "__pycache__" in dirpath:
            continue
        dirpath = dirpath.replace(cloud_eval_dir, '')
        listOfFiles += [os.path.join(dirpath, file) for file in filenames]
        for file_path in listOfFiles:
            framework_source_file = os.path.join(cloud_eval_dir, file_path)
            with open(framework_source_file, 'r', encoding="utf8") as file:
                try:
                    data = file.read()
                    data = data.replace('{name-your-service}', service_name)
                    data = data.replace('{name-your-artifacts}', service_name)
                    data = data.replace('generatedProjectName', project_name_no_underscore)
                    dirToCreate = ''
                    if main_dir:
                        dirToCreate = main_dir + '/cloud_eval/'
                    if not os.path.exists(dirToCreate):
                        os.makedirs(dirToCreate)
                    if "#cli_file_dest=[" in data:
                        start_ix = data.index("#cli_file_dest=[") + len("#cli_file_dest=[")
                        end_ix = data.index("]", start_ix)
                        destination_dir = data[start_ix:end_ix]
                        data = data[:data.index("#cli_file_dest=[")] + data[end_ix + 1:]
                        dirToCreate = main_dir + '/' + destination_dir

                    if not os.path.exists(dirToCreate + file_path):
                        f = open(dirToCreate + file_path, "w")
                        f.write(data)
                        f.close()
                        # Make sure that .sh files can be executed
                        if (dirToCreate + file_path).endswith('.sh'):
                            os.chmod(dirToCreate + file_path, 0o744)
                    else:
                        with open(dirToCreate + file_path, "r") as f:
                            existing_data = f.read()
                        if existing_data != data:
                            conflicted_file = f"{dirToCreate + file_path}"
                            conflicted_files.append((conflicted_file, str(framework_source_file)))

                except Exception as ex:
                    pass

    if conflicted_files:
        print("The following files already exists, so not overwriting them. "
              "However, notice it is different from the framework version in")
        for conflicted_file in conflicted_files:
            print(f"Existing: [{conflicted_file[0]}] \t\t Framework source file [{conflicted_file[1]}]")


def copy_batch_files(main_dir):
    """! Generate batch files needed for batch projects.
        Args:
            main_dir: A string describing the main working directory, on existing projects, it's empty.
    """

    current_folder = os.path.basename(os.getcwd())
    if main_dir != '':
        current_folder = main_dir
    service_name = current_folder.split('/')[-1]
    project_name_no_underscore = clean_existing_project_name(service_name)
    source_dir = os.path.join(__location__, 'dsframework/cli/tester/batch_files/')

    for (dir_path, dir_name, filenames) in os.walk(source_dir):
        if "__pycache__" in dir_path:
            continue

        dir_path = dir_path.replace(source_dir, '')
        file_list = [os.path.join(dir_path, file) for file in filenames]
        for file_path in file_list:
            file_suffix = Path(file_path).suffix
            with open(os.path.join(source_dir, file_path), 'r', encoding="utf8") as file:
                try:
                    data = file.read()
                    data = data.replace('{name-your-service}', service_name)
                    data = data.replace('generatedProjectName', project_name_no_underscore)

                    if main_dir == '':
                        target_dir = 'batch_files/'
                        dir_to_create = target_dir
                    else:
                        target_dir = current_folder + '/batch_files/'
                        dir_to_create = target_dir

                    if dir_path != "":
                        dir_to_create = target_dir + dir_path + '/'

                    if not os.path.exists(dir_to_create):
                        os.makedirs(dir_to_create)
                    if not os.path.exists(target_dir + file_path):
                        f = open(target_dir + file_path, "w")
                        f.write(data)
                        f.close()
                        if file_suffix == '.sh':
                            os.chmod(target_dir + file_path, 0o744)
                except Exception as ex:
                    if file_suffix == '.jar':
                        shutil.copy(os.path.join(source_dir, file_path), f'{dir_to_create}')
                    else:
                        print(f'Error when trying to open: {file_path}: {ex}')
                    pass


def copy_trainer_files():
    """! This function will create a trainer environment."""

    project_name = os.getcwd().split('/')[-1]
    trainer_main_folder = 'trainer/'
    trainer_plwrapper_folder = 'trainer/pl_wrapper/'

    trainer_file_list = ["trainer/pl_wrapper/custom_dataset",
                         "trainer/pl_wrapper/data_module",
                         "trainer/pl_wrapper/iterable_dataset",
                         "trainer/pl_wrapper/network_module",
                         "trainer/pl_wrapper/plmodel",
                         "trainer/comparison",
                         "trainer/config",
                         "trainer/data",
                         "trainer/dataset_preparation",
                         "trainer/main",
                         "trainer/model",
                         "trainer/test",
                         "trainer/train"]

    if not os.path.exists(trainer_main_folder):
        os.mkdir(trainer_main_folder)

    if not os.path.exists(trainer_plwrapper_folder):
        os.mkdir(trainer_plwrapper_folder)

    for filename_path in trainer_file_list:
        file_only = filename_path.split('/')[-1]
        parent_folder = filename_path.split('/')[-2]

        dst_folder = ''
        if parent_folder == 'pl_wrapper':
            dst_folder = trainer_plwrapper_folder

        data = read_template_file(filename_path)

        clean_project_name = clean_existing_project_name(project_name)
        filename_no_underscore = to_capitalize_no_underscore(file_only)

        class_name = clean_project_name + filename_no_underscore

        data = data.replace('generatedClass', class_name)
        data = data.replace('generatedProjectName', clean_project_name)

        new_file = trainer_main_folder + file_only + ".py"
        if dst_folder != '':
            new_file = dst_folder + file_only + ".py"

        if not os.path.isfile(new_file):
            create_file(new_file, data)
        else:
            print(f'File \'{new_file}\' exists, skipping.')

    source_path = os.path.join(__location__, 'dsframework/cli/', 'trainer/requirements_trainer.txt')
    shutil.copyfile(source_path, trainer_main_folder + '/requirements_trainer.txt')

    print("Creating trainer environment done !")


def copy_documentation_files(target_dir):
    """! Copy documentation folder to newly created project.

    Args:
        target_dir: Newly created project location
    """
    source_dir = os.path.join(__location__, 'dsframework/documentation/')
    shutil.copytree(source_dir, target_dir, ignore=shutil.ignore_patterns('dsf_doxy*'))


def update_doc_configuration(project_name, target_dir):
    """! Update project doxygen configuration file, adding project name.

    Args:
        project_name: Newly created project name
        target_dir: Configuration file location.
    """
    project_name_no_underscore = to_capitalize_no_underscore(project_name)
    source_dir = os.path.join(target_dir, 'project_doxyfile')
    data = read_file(source_dir)
    data = data.replace('generatedProjectName', project_name_no_underscore)
    save_file(source_dir, data)


@cli.command()
def install_trainer_packages():
    """! Install trainer packages."""
    command = 'pip install -r trainer/requirements_trainer.txt'
    subprocess.call(command, shell=True)


@cli.command()
def install_documentation_requirements():
    """! Install documentation doxygen."""

    doxygen_found = True

    try:  # Check for brew installation
        command = 'brew -v'
        res = subprocess.check_output(command, shell=True)
        brew_found = 'Homebrew' in str(res)

        if brew_found:
            try:  # Check for doxygen installation (if brew found)
                command = 'brew ls --versions doxygen'
                res = subprocess.check_output(command, shell=True)
                doxygen_found = 'doxygen' in str(res)

                if doxygen_found:
                    print('Doxygen already installed.')

            except subprocess.CalledProcessError as e1:
                doxygen_found = False

    except subprocess.CalledProcessError as e2:
        brew_found = False

    if not doxygen_found or not brew_found:
        installation_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'dsframework', 'documentation',
                                         'install_doxygen.sh')
        if not os.path.exists(installation_file):
            print('Installation script not found install_doxygen.sh')
            return
        is_mac_linux = (platform.system() == 'Darwin' or platform.system() == 'Linux')

        if is_mac_linux:
            subprocess.call(f'chmod +x {installation_file}', shell=True)

        subprocess.call(installation_file, shell=True)


def to_capitalize_no_underscore(text):
    """! Capitalize each word with no underscore
     for example : Moving from test_name -> TestName"""
    return ''.join(elem.capitalize() for elem in text.split('_'))


def clean_existing_project_name(text):
    no_underscore_text = to_capitalize_no_underscore(text)
    return ''.join(elem.capitalize() for elem in no_underscore_text.split('-'))


def let_user_pick(options):
    """! Let the user pick a number from couple of options
    @verbatim
    Args:
        options  : couple of options for user decision.
    @endverbatim
    """
    print("Please choose:")
    for idx, element in enumerate(options):
        print("{}) {}".format(idx + 1, element))
    i = input("Enter number: ")
    try:
        if 0 < int(i) <= len(options):
            return int(i) - 1
    except:
        pass
    return None


@cli.command()
def deploy_service_to_gae():
    """! Run deployment scripts in order to deploy the current project to GAE.
   """
    # Setup paths to script files
    setup_file_path = f"{__proj_root__}/deploy/setup_deploy_params.sh"
    framework_folder = os.path.dirname(os.path.abspath(__file__))

    # Check that we have a setup file. If not, copy it.
    if not os.path.exists(setup_file_path):
        print(f"file {setup_file_path} doesn't exist. Copying. Please fill relevant parameters and run again")
        shutil.copyfile(f"{framework_folder}/dsframework/cli/tester/deploy_files/deploy/setup_deploy_params.sh",
                        setup_file_path)
        return

    # Prepare commands
    setup_command = f"source {setup_file_path}"
    build_command = f"{framework_folder}/dsframework/base/deploy_scripts/build_and_register_image_in_devland.sh"
    deploy_command = f"{framework_folder}/dsframework/base/deploy_scripts/gae/deploy-gae-devland.sh"
    full_command = f"{setup_command};{build_command};{deploy_command}"

    # Execute
    subprocess.call(full_command, shell=True)
    # subprocess.call(setup_command, shell=True)
    # subprocess.call(build_command, shell=True)
    # subprocess.call(deploy_command, shell=True)


if __name__ == '__main__':
    cli(prog_name='cli')



import redis
import logging
from typing import List, Any
from redis import RedisError
from dsframework.base.server.cache.cache import CacheItem, Cache

logger = logging.getLogger(__name__)


class RedisCache(Cache):
    """!The Redis implementation of the Cache interface. Uses a StrictRedis as the underlying Redis client.

    Attributes:
        host: the Redis server connection host (localhost by default)
        port: the Redis server connection port (6379 by default)
        ttl_seconds: time-to-live - how long before items are evicted from the cache (30 days by default)
    """

    def __init__(self, host: str = 'localhost', port: int = 6379,
                 ttl_seconds: int = 60 * 60 * 24 * 30, socket_timeout: int = 5) -> None:
        self.redis_client: redis.client.Redis = redis.StrictRedis(host=host, port=port, socket_timeout=socket_timeout)
        self.redis_connected: bool = False
        self.host: str = host
        self.port: int = port
        self.ttl_seconds: int = ttl_seconds  # time-to-live: items expire from cache after 30 days by default

    def init(self):
        """!Initialization for the cache in any manner (for example: connect)"""
        self.connect()

    # noinspection PyBroadException
    def connect(self):
        """!Connects the client to the Redis server"""
        logger.info("Connecting to Redis...")
        self.redis_client: redis.client.Redis = redis.StrictRedis(host=self.host, port=self.port, socket_timeout=5)
        self.redis_connected = False
        try:
            self.redis_client.ping()
            self.redis_connected = True
            logger.info('Connected to Redis successfully')
        except Exception:
            logger.exception('ERROR Failed connecting to Redis')
            return

    def status(self) -> bool:
        """!Indicates if the cache client is ready for work or not.

        :return: True when the cache client is ready for work, False otherwise
        """
        return self.redis_connected

    def get(self, keys: List[str]) -> List[Any]:
        """!Retrieves one or more items from the cache

        :param keys: the list of keys to retrieve items for
        :return: a list of one or more items corresponding to the given list of keys
        """
        retrieved_items = []
        if not self.redis_connected:
            return []
        try:
            retrieved_items = self.redis_client.mget(keys)  # retrieve multiple keys
        except RedisError as ex:
            exception_type = type(ex).__name__
            logger.error(f'ERROR Failed retrieving items from Redis cache - {keys} - {exception_type}')
            if exception_type == 'ConnectionError':
                self.redis_connected = False
        return retrieved_items

    def set(self, cache_items: List[CacheItem]):
        """!Stores one or more items in the cache

        :param cache_items: the list of CacheItems (key:value pairs) to be stored
        """
        if not self.redis_connected:
            return
        try:
            pipe = self.redis_client.pipeline(transaction=False)
            for cache_item in cache_items:
                pipe.set(name=cache_item.key, value=cache_item.value, ex=self.ttl_seconds)
            pipe.execute()
        except RedisError as ex:
            exception_type = type(ex).__name__
            logger.error(f'ERROR Failed storing items in Redis cache - {cache_items} - {exception_type}')
            if exception_type == 'ConnectionError':
                self.redis_connected = False



import os
import sys
import argparse
from typing import List
from google.cloud.dataproc_v1.types.workflow_templates import TemplateParameter
from google.cloud.dataproc_v1.types import OrderedJob
from dsframework.base.batch.workflow_base import ZIDS_Workflow


class Workflow(ZIDS_Workflow):

    def __init__(self, workflow_name: str = ""):
        """! Loads config file and initializes parameters.
        """
        self.default_jars = ['gs://spark-lib/bigquery/spark-bigquery-latest_2.12.jar']
        main_project_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "../.."))

        super().__init__(workflow_name)

        # Add general packages, as well as workflow specific ones, if available
        if not self.config["use_only_wf_packages"]:
            self.configure_pip_packages(main_project_path + '/requirements_docker.txt')

        curr_wf_path = os.path.abspath(os.path.join(os.path.dirname(__file__), f"{self.wf_name}_workflow"))
        self.configure_pip_packages(curr_wf_path + f"/requirements_{self.wf_name}.txt", append=True)

        # Option for another build type - dependent on setup.py
        self.whl_file_name = f"{self.project_name}_{self.wf_name}.zip"

        # Initialize spark properties.
        self.spark_properties = self.config["spark_properties"]

    def create_job(self, stage_name, prerequisite_list=None) -> OrderedJob():
        """! Create a job from a given stage

            Args:
                stage_name: Name of the stage to create
                prerequisite_list: list of stage id's that are mandatory to run before current stage

            Returns:
                An Ordered Job, to be inserted into the workflow template
        """
        stage_dir = f'batch_files/stages/{stage_name}'
        python_files = [self.whl_file_name]
        archive_files = [self.whl_file_name]

        # Merge both basic configuration and stage specific ones
        if stage_name in self.spark_properties:
            stage_spark_properties = {**self.spark_properties['basic'], **self.spark_properties[stage_name]}
        else:
            stage_spark_properties = self.spark_properties['basic']

        # Convert to desired "List of lists" format
        spark_properties_list = [[key, value] for key, value in list(stage_spark_properties.items())]

        stage_job = self.create_ordered_job(stage_dir, spark_properties_list, python_files, archive_files)

        if prerequisite_list is not None:
            stage_job.prerequisite_step_ids = [f"{self.project_name}-{stage}" for stage in prerequisite_list]

        return stage_job

    def create_jobs(self) -> List[OrderedJob]:
        """! Generates a job for each of the workflow stages
            Returns:
                A List of OrderedJob, to be inserted into the workflow template
        """

        ordered_jobs = [self.create_job(stage['stage_name'], stage['prerequisites']) for stage in self.wf_stages]

        return ordered_jobs

    def finalize_jobs(self, workflow_template):
        """! Finalize the jobs, introduce additional parameters, etc.
            Args:
                workflow_template: Workflow templates. Includes jobs that we wish to modify
        """

        # Make sure to change jobs according to specific configuration
        if self.config['specific_jobs_to_run'] and len(self.config['specific_jobs_to_run']):
            specific_jobs_to_run = []
            for job in workflow_template.jobs:
                if job.step_id in self.config['specific_jobs_to_run']:
                    specific_jobs_to_run.append(job)
                    if self.config['force_remove_all_prerequisite']:
                        job.prerequisite_step_ids = []
                    if len(job.prerequisite_step_ids) and not all(
                            elem in specific_jobs_to_run for elem in job.prerequisite_step_ids):
                        raise Exception("job depends on another job not in specific_jobs_to_run list")
            workflow_template.jobs = specific_jobs_to_run

        # Set start and end dates from Airflow composer
        workflow_template.parameters = []

        start_date_param = TemplateParameter()
        start_date_param.description = "start date from airflow composer"
        start_date_param.name = "START_DATE"

        end_date_param = TemplateParameter()
        end_date_param.description = "end date from airflow composer"
        end_date_param.name = "END_DATE"

        extra_param = TemplateParameter()
        extra_param.description = "allow adding extra parameters"
        extra_param.name = "PARAMS"

        for job in workflow_template.jobs:
            start_date_param.fields.append("jobs['" + job.step_id + "'].pysparkJob.args[1]")
            end_date_param.fields.append("jobs['" + job.step_id + "'].pysparkJob.args[2]")
            extra_param.fields.append("jobs['" + job.step_id + "'].pysparkJob.args[3]")

        workflow_template.parameters.append(start_date_param)
        workflow_template.parameters.append(end_date_param)
        workflow_template.parameters.append(extra_param)

    def create_template_structure(self):
        """! Create the workflow template, and return it for external usage.
            Returns:
                workflow_template: A ready to use template, WorkflowTemplate() class
        """
        # Define template structure, along with cluster requirements
        workflow_template = self.build_template_structure()

        # Define jobs that this workflow will run on a given compute cluster
        workflow_template.jobs = self.create_jobs()

        # Finalize job parameters
        self.finalize_jobs(workflow_template)

        return workflow_template

    def create_workflow_template(self, create_yaml: bool = False):
        """! Create the workflow template, along with all relevant stages
            Args:
                create_yaml: Boolean, deciding whether we should create a yaml file out of the template
        """

        # Check if a cluster exists. If it does, we will work with it. If not, we'll use the managed cluster from the
        # configuration file
        self.check_cluster()

        # Define template structure
        workflow_template = self.create_template_structure()

        print(f'{workflow_template}')

        # Make sure that the template does not exist
        self.check_template()

        # If needed, create a .yaml file, don't use Cloud API
        if create_yaml:
            self.create_yaml_file(workflow_template)

        else:
            # Create template on Dataproc by using Cloud API
            self.create_template(workflow_template)


if __name__ == "__main__":
    """! Triggers the creation of the workflow template, starting from
    basic configuration, defining and creating jobs from the existing stages, 
    and up to creating and instantiating the workflow, if needed.
        Args:
        System argument 1 - Workflow name
        System argument 2 - Action to be performed on a template (create, delete, instantiate)
    """

    parser = argparse.ArgumentParser(description='Run various workflow related commands')
    parser.add_argument('--wf-name', type=str, default='primary',
                        help='workflow name which will be the base workflow for related actions')
    parser.add_argument('--action', type=str, default='create',
                        help='Workflow action (create / create_yaml / delete / instantiate')

    args = parser.parse_args()

    template_action = args.action

    workflow = Workflow(args.wf_name)

    if template_action == 'create':
        workflow.create_workflow_template()

    if template_action == 'create_yaml':
        workflow.create_workflow_template(create_yaml=True)

    # At this point, a template is either previously or recently created
    template = workflow.get_workflow_template()

    if template_action == 'delete' and template:
        workflow.delete_workflow_template(template)

    if template_action == 'instantiate' and template:
        workflow.instantiate_template(template)



from dsframework.base.trainer.model import ZIDSTrainerModel
from trainer.pl_wrapper.plmodel import generatedProjectNamePlmodel

##
# @file
# @brief generatedClass class, defines model class.


class generatedClass(ZIDSTrainerModel):
    """Model class - defines the network, loss function etc."""

    def __init__(self, model_config, metrics_config, trainer_config, save_path):
        super().__init__(model_config, metrics_config, trainer_config, save_path)

    def define_model(self):
        """! Implement model class in the format you use in the fit/train method.
        Create a new model or load from pretrained model.

        Needs to be implemented.

        Example (pytorch lightning):
            self.model = generatedProjectNamePlmodel(
                model_config=self.model_config,
                metrics_config=self.metrics_config,
                trainer_config=self.trainer_config,
                save_path=self.save_path
            )
        """
        if self.model is None:
            raise Exception('Implement model in model.py')

    def get_model(self):
        """Returns model class"""
        return self.model



"""! @brief Get Data stage, for obtaining data from a generic source, such as a database."""
import argparse
import os
import sys
from typing import Any
from json import loads as json_loads

from dsframework.base.batch.pipeline_stage_base import ZIDS_BatchPipelineStage
from trainer.main import Main
from trainer.config import config as trainer_config

##
# @file
# @brief Stage main class, implements ZIDS_Stage base class.
class RunTrainerStage(ZIDS_BatchPipelineStage):
    """! Stage class

    Implement a stage that will later be converted to an executable job in a specific workflow.
    """

    def __init__(self, stage_config):
        """! The Stage class (generatedStageName) initializer.
        Base class will load basic configuration parameters, additional fields should be added here

            Args:
                stage_config : Configuration dictionary, loaded from configuration file.
        """

        ##
        # @hidecallgraph @hidecallergraph
        print(f"Initializing Stage: {self.get_name()}")

        self.wf_name = "train"
        super().__init__(stage_config)

    def get_name(self):
        """! Get the stage name
        """
        return self.__class__.__name__

    def main(self, **kwargs: Any):
        """! Executes the main functionality of the stage.

            Args:
                **kwargs : Whatever is needed for the stage to run properly.
        """
        trainer_main_class = Main(cfg=trainer_config, training_path=os.getcwd() + '/datasets',
                          validation_path=os.getcwd() + '/datasets')
        trainer_main_class.execute_trainer(force_save_path='/trainer_tmp')
        trainer_main_class.copy_to_gs(self.bucket_name, self.folder_path, self.project_id)

        print('Trainer stage completed.')


if __name__ == "__main__":
    """! Executes the stage by instantiating it and calling the main function.
    Set up argument condition according to the usage of the written stage

        Args:
            System argument 1 - Configuration file
            System argument 2 - Start date
            System argument 3 - End date, received from Airflow
            System argument 4 - Placeholder for future parameters
    """

    parser = argparse.ArgumentParser(description='Configure and run RunTrainerStage class')
    parser.add_argument('config', type=str,
                        help='Configuration file')
    parser.add_argument('start_date', type=str, default='',
                        help='Start date, received from Airflow')
    parser.add_argument('end_date', type=str, default='',
                        help='End date, received from Airflow')
    parser.add_argument('params', nargs=argparse.REMAINDER,
                        help='Placeholder for future parameters')
    args = parser.parse_args()

    # Configure stage class and run
    stage = RunTrainerStage(json_loads(args.config))
    try:
        stage.update_stage_params(args.start_date, args.end_date, args.params)
        stage.main()
    except Exception as e:
        raise Exception(f" Stage failed with error: {e}")



#!/usr/bin/env python
# coding: utf-8

"""! @brief ZIDS_Predictable base class."""

from typing import Any
from dsframework.base.common import ZIDS_Object


class ZIDS_Predictable(ZIDS_Object):
    """! ZIDS_Predictable base class for a single predictable object."""

    def __init__(self) -> None:
        """! Initializer for ZIDS_Predictable
        Initializes local class variables, those ZIDS_Predictable variables will be transferred from component to
        component in the pipeline as an input and output.
        """
        ##
        # @hidecallgraph @hidecallergraph
        super().__init__()
        self.name: str = ''
        self.input: Any = None
        self.target: Any = None
        self.pred: Any = None
        self.prob: float = -1.0
        self.forced_pred: Any = None
        self.forced_reason: str = 'Did Not Force'



class ZIDSTrainerModel:
    
    def __init__(self, model_config, metrics_config, trainer_config, save_path):
                
        self.model = None
        self.model_config = model_config
        self.metrics_config = metrics_config
        self.trainer_config = trainer_config
        self.save_path = save_path
        
        self.define_model()

    def define_model(self):
        """! Implement model class in the format you use in the fit/train method.
        Create a new model or load from pretrained model.

        Needs to be implemented.

        Example (pytorch lightning):
            self.model = generatedProjectNamePlmodel(
                model_config=self.model_config,
                metrics_config=self.metrics_config,
                trainer_config=self.trainer_config,
                save_path=self.save_path
            )
        """
        if self.model is None:
            raise Exception('Implement model in model.py')

    def get_model(self):
        return self.model



import os
from dsframework.base.trainer.pl_wrapper import *

##
# @file
# @brief config.py, holds all configurations in dictionary format.

## ModelCheckpoint - pytorch lightning auto save callback.
checkpoint_callback = ModelCheckpoint(
    save_top_k=1,
    save_last=True,
    monitor="val_loss",
    mode="min",
    filename="my-chkpnt-{epoch:02d}-{val_loss:.2f}",
)

## EarlyStopping - pytorch lightning early stopping callback.
early_stopping_callback = EarlyStopping(
    monitor='val_loss',
    patience=10,  # number of val checks with no improvement
    mode='max'
)

## Supported loggers by pytorch lightning:
## TensorBoardLogger, MLFlowLogger, NeptuneLogger, WandbLogger, CometLogger, CSVLogger.
# By default created the TensorBoardLogger and it is the only one tested.
# More information here: https://pytorch-lightning.readthedocs.io/en/stable/extensions/logging.html
pl_logger = TensorBoardLogger(save_dir=os.getcwd(), version=f'v', name='training_output')
device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

## Configurations, can add or remove parameters with two exception: metrics and the trainer_config is the exact parameters
## the pytorch lightning trainer class accepts, it is used: pl.Trainer(**config['trainer_config'])
config = {
    'model_config': {
        'learning_rate': 1e-3,
        'loss_function': F.cross_entropy,
        'batch_size': 32,
        'num_classes': None,
        'num_classes_conf_mat': None,
        'label_names': None,
        'num_workers': 1,
        'seed': 42,
        'save_last_model': False,
        'model_name': 'my_model.ckpt',
        'device': device,
        'num_warmup_steps': 0,
        'save_checkpoint_as_onnx': False,
        'average': 'micro'  # 'macro'
    },
    'trainer_config': {  # Do not change, training_config is dedicated to pytorch lighting trainer module and uses its exact params.
        'max_epochs': 1,
        'accelerator': None,  # 'cpu', 'gpu', 'tpu', 'ipu', 'auto', None
        'devices': None,  # number of accelerators: -1-all, 1, 3, 4...8, None, [1, 3]-gpu ids, None
        'strategy': None,  # 'dp', 'ddp', 'ddp2', 'ddp_spawn'
        'num_processes': None,  # number of cpus
        'enable_checkpointing': True,
        'default_root_dir': './',  # root dir to save checkpoints
        'callbacks': [checkpoint_callback],
        'enable_progress_bar': True,  # PRD - for production change to False
        'log_every_n_steps': 20,
        'logger': pl_logger,
        'num_sanity_val_steps': 0  # =0 will not perform sanity test.
    },
    'metrics': {  # works with pytorch lightning and will calculate and return the specified metrics.
        'precision': True,
        'accuracy': True,
        'recall': True,
        'f1': True,
        'confusion_matrix': True
    },
    'dataset_config': {
        'save_dataset': True,
        'train_ratio': 0.75,
        'validation_ratio': 0.15,
        'test_ratio': 0.10
    }
}



from os.path import exists
from typing import Optional, Union, Any
import csv
from dsframework.base.trainer.pl_wrapper import *


class ZIDS_PLDataModule(pl.LightningDataModule):
    """! This class inherits from LightningDataModule"""

    def __init__(self,
                 model_config=None,
                 train_set: Dataset = None,
                 val_set: Dataset = None,
                 test_set: Dataset = None):
        """! Define required parameters here."""
        super().__init__()

        self.train_val_len = 0
        self.train_val_set = train_set
        self.val_set = val_set
        self.train_set = None if val_set is None else train_set
        self.test_set = test_set
        self.predict_set = None
        self.batch_size = model_config['batch_size']
        self.num_workers = model_config['num_workers']

    def prepare_data(self):
        """! Define steps that should be done on only one GPU, like getting data."""
        pass

    def setup(self, stage: Optional[str] = None) -> None:
        """! Make assignments here (val/train/test split)
        called on every process in DDP
        Define steps that should be done on
        every GPU, like splitting data, applying
        transform etc.


        Example:
            @code
            # Assign train/val datasets for use in dataloaders:
            if stage in (None, "fit"):
                mnist_full = MNIST(self.data_dir, train=True, transform=self.transform)
                self.ds_train, self.ds_val = random_split(mnist_full, [55000, 5000])

            # Assign test dataset for use in dataloader(s):
            if stage in (None, "test"):
                self.ds_test = MNIST(self.data_dir, train=False, transform=self.transform)

            @endcode
        """
        pass

    def train_dataloader(self) -> TRAIN_DATALOADERS:
        """! Return DataLoader for Training Data here

        Example:
            @code
            train_split = Dataset(...)
            return DataLoader(train_split)
            @endcode
        """
        return DataLoader(self.train_set, batch_size=self.batch_size, num_workers=self.num_workers)

    def val_dataloader(self) -> EVAL_DATALOADERS:
        """! Return DataLoader for data validation.

        Example:
            @code
            val_split = Dataset(...)
            return DataLoader(val_split)
            @endcode
        """
        return DataLoader(self.val_set, batch_size=self.batch_size, num_workers=self.num_workers)

    def test_dataloader(self) -> EVAL_DATALOADERS:
        """! Return DataLoader for data testing.

        Example:
            @code
            val_split = Dataset(...)
            return DataLoader(test_split)
            @endcode
        """
        return DataLoader(self.test_set, batch_size=self.batch_size, num_workers=self.num_workers)

    def predict_dataloader(self) -> EVAL_DATALOADERS:
        return DataLoader(self.predict_set, batch_size=self.batch_size)

    def teardown(self, stage: Optional[str] = None) -> None:
        """! Called at the end of fit(train + validate), validate, test, or predict.
            called on every process in DDP
        """
        pass

    def load_dataset(self, file_path, file_type):
        """! Loads an external dataset, NOT FINISHED YET."""

        if file_type == 'csv':
            with open(file_path) as csv_file:
                self.train_val_set = list(csv.reader(csv_file))

        return self.train_val_set

    def parse_dataset(self, remove_header) -> Union[Any, Any]:
        """! Parse your data here, it needs to return x, y.
            Returns:
                x: dataset
                y: target

        Example:
            @code{.py}
            if remove_header:
                self.dataset = self.dataset[1:]

            x = np.array(self.dataset, dtype=np.float32)[:, :-1]
            y = np.array(self.dataset, dtype=np.float32)[:, -1].reshape(-1, 1)

            return x, y
            @endcode
        """
        raise Exception("Not implemented exception (parse_dataset).")



import sys
import argparse
from datetime import datetime, timedelta
from textwrap import dedent

from dsframework.base.batch.dag_base import ZIDS_Dag


class Dag(ZIDS_Dag):
    def __init__(self, workflow_name: str = ""):
        """! Loads config file and initializes parameters.
        Please make sure to determine the start date according to your project needs.
        """

        # dag params from airflow 1 - https://airflow.apache.org/docs/apache-airflow/1.10.3/macros.html
        # dag params from airflow 2 - https://airflow.apache.org/docs/apache-airflow/stable/templates-ref.html
        super().__init__(workflow_name)

        self.file_name = self.project_name + "_" + self.wf_name + "_dag.py"

        # This setting will give the dag an execution date of the day before its creation.
        self.dag_start_delay = 1

    def create_dag(self):
        """! Dag creation.
        This function creates the python file which contains the DAG itself.
        """

        with open(self.file_name, 'w') as dag_text_file:
            dag_text_file.write(dedent(self.get_basic_dag_code()))
        print(f"DAG file created: {self.file_name}")


if __name__ == "__main__":
    """! Triggers the creation of the workflow dag, which will instantiate a
    workflow template into an actual workflow, executing it.
    Through this module, we can create a dag file which contains instructions, regarding how to execute the 
    workflow template, and we can also import / remove it into / from an active GCP composer (Only for dev mode)
        Args:
            System argument 1 - Workflow to be handled
            System argument 2 - Action to be performed on a dag (create, import, delete)
            
    """
    parser = argparse.ArgumentParser(description='Run various DAG related commands')
    parser.add_argument('--wf-name',
                        type=str,
                        default='primary',
                        help='workflow name which will be the DAG base workflow')
    parser.add_argument('--action',
                        type=str,
                        default='create',
                        help='DAG action (create / import / delete')

    args = parser.parse_args()

    dag = Dag(args.wf_name)

    if args.action == 'create':
        dag.create_dag()

    elif args.action == 'import':
        dag.import_dag()

    elif args.action == 'delete':
        dag.delete_dag()



class generatedClass:

    def __init__(self, dataset_path='', save_datasets=False, save_path='', dataset_config=None):
        self.dataset_path = dataset_path
        self.save_datasets = save_datasets
        self.save_path = save_path
        self.dataset_config = dataset_config
        self.dataset_df = self._read_dataset(dataset_path)

    def prepare_datasets(self):
        """ Do here all dataset preparation for splitting.

        Example:

            if self.save_datasets:
                with open(self.save_path + '/train_dataset.pickle', 'wb') as pickle_file:
                    pickle.dump(x_train_data, pickle_file)
                with open(self.save_path + '/test_dataset.pickle', 'wb') as pickle_file:
                    pickle.dump(x_test_data, pickle_file)
                with open(self.save_path + '/val_dataset.pickle', 'wb') as pickle_file:
                    pickle.dump(x_val_data, pickle_file)
                return
                    self.save_path + '/train_dataset.pickle',
                    self.save_path + '/val_dataset.pickle',
                    self.save_path + '/test_dataset.pickle'

            else:
                return pickle.dumps(x_train_data), pickle.dumps(x_val_data), pickle.dumps(x_test_data)


        """

        train_dataset_path = self.save_path + '/train_dataset.pickle'
        val_dataset_path = self.save_path + '/val_dataset.pickle'
        test_dataset_path = self.save_path + '/test_dataset.pickle'

        return train_dataset_path, val_dataset_path, test_dataset_path

    def _read_dataset(self, dataset_path):
        """Reads the full dataset.

        Example:
            dataset_df = pd.read_csv(self.dataset_path)
        """

        dataset_df = None

        return dataset_df




from dsframework.base.tester.evaluator_logic import ZIDS_Evaluator


class generatedClass(ZIDS_Evaluator):
    def __init__(self):
        """! generatedClass initializer """
        ##
        # @hidecallgraph @hidecallergraph
        super().__init__()

    def get_confusion_matrix(self, true_label, prediction, index):
        """! Model evaluation function, to be implemented per project in evaluator.py file

        Important:
            This method is not implemented, please override to use its functionality.
        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def prepare_row_result(self, ground_truth, model_predication, conf_mat):
        """! Preparation for each row before calling each attribute get_confusion_matrix func

        Args:
            ground_truth:str : DataSet ground truth string
            model_predication:str : Model predication string
            conf_mat:dict : Confusion matrix for one row on the DataSet
        Important:
            This method is not implemented, please override to use its functionality.
        Raises:
            NotImplementedError
        """
        raise NotImplementedError



#!/usr/bin/env python
# coding: utf-8

from pydantic.main import BaseModel
##
# @file
# @brief Setting the output schema of the pipeline / service.
#        What are the expected variables to output?


class generatedClass(BaseModel):
    """! Examples:
    @code
    class <my-new-project>Outputs(BaseModel):
        is_valid: int = 0
        prob: float = 0.5
        version: str
    @endcode
    @code
    class <my-new-project>Outputs(BaseModel):
        pred: bool
        prob: float
        version: str = ''
    @endcode
    """
    pass



from typing import Any


class ZIDSNetwork:
    """! Base class for neural network definition:

    Example:
        @code
        model = nn.Sequential(
            nn.Linear(28 * 28, 64),
            nn.ReLU(),
            nn.Linear(64, 3),
            nn.Linear(3, 64),
            nn.ReLU(),
            nn.Linear(64, 28 * 28)
        )
        @endcode

    for example:
        @code
            model = nn.Sequential(
                nn.Linear(28 * 28, 64),
                nn.ReLU(),
                nn.Linear(64, 3),
            )
        @endcode

    Another example:
        @code
        self.bert = BertModel.from_pretrained("bert-base-cased")
        @endcode
    """

    model = None

    def __init__(self, model=None):

        self.model = model

    @staticmethod
    def load_define_model():
        """ Load/Define model here, getting called from plmodel.py


        Example:
            @code
            model = nn.Sequential(
                nn.Linear(28 * 28, 64),
                nn.ReLU(),
                nn.Linear(64, 3),
                nn.Linear(3, 64),
                nn.ReLU(),
                nn.Linear(64, 28 * 28)
            )
            @endcode

        Another example:
            @code
                model = nn.Sequential(
                    nn.Linear(28 * 28, 64),
                    nn.ReLU(),
                    nn.Linear(64, 3),
                )
            @endcode

        Another example:
            @code
                self.bert = BertModel.from_pretrained("bert-base-cased")
            @endcode

        Returns:

        """
        model = None
        tokenizer = None

        if model is None:
            raise Exception("load_model() not defined. please define in network_module.py")

        return model, tokenizer



class ZIDSTrainClass:

    train_dataset = None
    validation_dataset = None
    trainer = None
    model = None
    data_module = None
    trainer_config = None
    model_config = None
    save_path = None

    def __init__(self, trainer_config=None, model_config=None, train_dataset=None, validation_dataset=None,
                 model=None, data_module=None, save_path=''):
        self.train_dataset = train_dataset
        self.validation_dataset = validation_dataset
        self.model = model
        self.data_module = data_module
        self.trainer_config = trainer_config
        self.model_config = model_config
        self.trainer = None  # trainer instance
        self.save_path = save_path

    def execute(self):
        """! Training process.

        for example:
            self.trainer = pl.Trainer(**self.config)
            self.trainer.fit(self.model, self.data_module)
            self.trainer.save_checkpoint("my_model.ckpt")
        """
        pass

    def get_trainer(self):
        return self.trainer



import os
from os.path import exists
from pathlib import Path
import csv
import shutil
from google.cloud import storage
from pytorch_lightning.loggers import LightningLoggerBase, TensorBoardLogger
from trainer.comparison import generatedProjectNameComparison
from trainer.config import config
from trainer.data import generatedProjectNameData
from trainer.model import generatedProjectNameModel
from trainer.test import generatedProjectNameTest
from trainer.train import generatedProjectNameTrain
from tester.general_tester import generatedProjectNameGeneralTester
from trainer.dataset_preparation import generatedProjectNameDatasetPreparation

##
# @file
# @brief Main class.


class Main:
    """! Trainer main class, this is where all training components orchestrate.

    Its main job is to run:
    - training - train the model using train dataset and validation dataset.
    - testing - running model against test dataset.
    - eval - running the same test dataset

    Important node:
    It includes a split() method, which needs to be implemented. This is where the dataset gets split to
    train, validation and test datasets.

    """
    model_config = None
    trainer_config = None

    def __init__(self,
                 cfg=None,
                 training_path=None,
                 validation_path=None,
                 test_dataset_path=None,
                 force_dataset_split=False):

        self.training_path = training_path
        self.validation_path = validation_path
        self.test_dataset_path = test_dataset_path
        self.force_dataset_split = force_dataset_split

        """! Trainer main initializer."""
        self.save_path = None
        self.train_class = None
        self.model_class = None
        self.data_class = None
        self.data_class_test = None
        self.test_class = None
        self.model_config = cfg['model_config']
        self.trainer_config = cfg['trainer_config']
        self.metrics_config = cfg['metrics']
        self.dataset_config = cfg['dataset_config']
        self.dataset = None
        self.dataset_train = None
        self.dataset_validation = None
        self.dataset_test = None
        self.eval_res = None

        if (not training_path or not validation_path) and not test_dataset_path:
            raise 'Please provide paths for training and validation or test datasets in Main.'

        if not (self.test_dataset_path and exists(self.test_dataset_path)):
            if self.force_dataset_split or not self.training_path or not exists(self.training_path):
                self.prepare_datasets()

    def prepare_datasets(self):
        """! Implement this method to prepare/split the dataset to train/val/test datasets:
        For the trainer supply: dataset_train, self.dataset_validation
        For test supply: self.dataset_test

        Example:
            self.dataset = datasets.MNIST(self.dataset_path, download=True, train=True,
                                      transform=transforms.ToTensor())
            self.dataset_train, self.dataset_validation = random_split(self.dataset, [55000, 5000])
            self.dataset_test = datasets.MNIST(self.dataset_path, download=True, train=False,
                                           transform=transforms.ToTensor())

        Another example:
            dataset_prep = generatedProjectNameDatasetPreparation()

            train_dataset_path,
            val_dataset_path,
            test_dataset_path = dataset_prep.prepare_datasets(self.dataset_path, save_datasets=self.save_split_datasets,
                save_path=self.dataset_path, config=self.dataset_config)

        """

        #  raise 'Dataset Main.prepare_datasets() not implemented exception.'
        pass

    def save_dataset(self, dataset_path):
        """! Saves a copy of the dataset in /training_output/v1 folder zipped.

        In a later stage the trained model is created in the same folder, so we have a pair of the model and the
        dataset used to create it.
        """
        if not self.dataset_config['save_dataset']:
            return

        if os.path.isdir(dataset_path):
            if os.listdir(dataset_path):
                shutil.make_archive(os.path.join(self.save_path, 'dataset'), 'zip', dataset_path)

    def execute_trainer(self, force_save_path: str = ''):
        """ Executes training process. it initializes the main 3 trainer modules:

        generatedProjectNameData - Class for holding datasets and functionality related to dataset handling
        generatedProjectNameModel - Network definition and all functionality related to creating the model.
        generatedProjectNameTrain - Execute trainer and all trainer related functionality.
        """

        self.save_path = self._save_path(output_path=force_save_path)

        self.model_class = generatedProjectNameModel(
            model_config=self.model_config,
            metrics_config=self.metrics_config,
            trainer_config=self.trainer_config,
            save_path=self.save_path)

        self.data_class = generatedProjectNameData(
            model_config=self.model_config,
            load_datasets=True,
            training_path=self.training_path,
            validation_path=self.validation_path,
            tokenizer=self.model_class.model.tokenizer)

        self.train_class = generatedProjectNameTrain(
            trainer_config=self.trainer_config,
            model_config=self.model_config,
            data_module=self.data_class.get_data_module(),
            model=self.model_class.get_model(),
            save_path=self.save_path)

        self.train_class.execute()

        self.save_dataset(self.training_path)

    def execute_test(self, model_path):
        """ Executes test process. runs the model against a test dataset.

        Initializes:
            generatedProjectNameData - Created with test dataset.
            generatedProjectNameTest - Loads model and executes test.

        Args:
            model_path: Path to saved model
        Returns:
            mmodel_test_results: Metrics - defined in config.
        """

        self.save_path = self._save_path()

        self.test_class = generatedProjectNameTest(
            model_path=model_path,
            test_dataset_path=self.test_dataset_path,
            trainer_config=self.trainer_config,
            model_config=self.model_config,
            metrics_config=self.metrics_config,
            save_path=self.save_path
        )

        model_test_results = self.test_class.execute_test()

        return model_test_results

    def execute_eval(self):
        """ Executes evaluation process on an existing model against a test dataset.

        Notes:
            Test dataset must be in Pandas dataframe format.

        Returns:
            eval_res - Evaluation metrics result.

        Result format example:
            self.eval_res = {'precision': 1.1, 'accuracy': 1.2, 'recall': 1.3, 'f1': 1.4,
                            'confusion_matrix': {'tp': 0, 'tn': 0, 'fp': 0, 'fn': 0}}
        """

        model_tester = generatedProjectNameGeneralTester()

        try:
            output_path = model_tester.test_from_dataset(self.dataset_test, 1)
            self.eval_res = self.get_eval_results(output_path)

        except Exception as e:
            print(f'Could not start evaluation process: {e}')

        return self.eval_res

    def get_eval_results(self, output_path):
        """Reads evaluation results from path and returns it.

        Args:
            output_path - Location of eval results

        Returns:
            evaluation results - In dictionary format.

        """
        if exists(output_path):
            return csv.DictReader(open(output_path))
        else:
            return None

    def _save_path(self, output_path=None):
        """Returns the save path, based on the next version number."""

        if not output_path:
            output_path = os.path.dirname(os.path.realpath(__file__))

        training_output_path = os.path.join(output_path, 'training_output')

        if isinstance(self.trainer_config['logger'], TensorBoardLogger):
            version_num = self._get_next_folder_num(training_output_path)
            save_path = self._get_save_path(create=False, curr_folder=training_output_path)
            self.trainer_config['logger'] = TensorBoardLogger(save_dir=output_path, version=f'v{version_num}',
                                                              name='training_output')
        else:
            save_path = self._get_save_path(create=True, curr_folder=training_output_path)

        return save_path

    def _get_save_path(self, create=False, curr_folder=None):
        """Create the next version folder (v1, v2, v3 ... vn) inside training_output folder.

        Returns:
            save_path - The newly created folder path.
        """

        if not curr_folder:
            curr_folder = os.path.dirname(os.path.realpath(__file__))

        if not os.path.exists(curr_folder):
            os.makedirs(curr_folder)

        new_version = self._get_next_folder_num(curr_folder)

        save_path = os.path.join(curr_folder, f'v{new_version}')

        if create:
            os.makedirs(save_path)

        return save_path

    def _get_next_folder_num(self, path) -> int:
        """ Checks input folder for the next version number: v1, v2, v3 .... vn

        Args:
            path: Path of the output folder.

        Returns:
            an integer indicating the next version number.
        """
        if not os.path.exists(path):
            os.makedirs(path)

        all_folders = [file for file in os.listdir(path) if os.path.isdir(os.path.join(path, file))]

        new_version = 1

        if all_folders:
            num_list = [int(folder.replace('v', '')) for folder in all_folders if folder.replace('v', '').isdigit()]
            num_list.sort()
            if num_list:
                new_version = num_list[-1] + 1

        return new_version

    def copy_to_gs(self, bucket_name, folder_path, project_id):
        """Copies training results folder to bucket
        Args:
            bucket_name: GCP bucket name, for example: us-east1-ds-airflow-1-dev-f63d7361-bucket
            folder_path: GCP bucket folder, save to path
            project_id: GCP Project name, for example: dozi-data-science-research-1
        """
        local_path = self.save_path
        storage_client = storage.Client()
        bucket = storage_client.bucket(bucket_name)

        print('local_path: ', local_path)

        if not os.path.isdir(local_path):
            raise Exception(f'Results folder not found: {local_path}')

        for root, dirs, files in os.walk(local_path):
            for name in files:
                local_file = os.path.join(root, name)
                print('local_file: ', local_file)

                filename = local_file.split('/')[-1]
                blob = bucket.blob(folder_path + local_file)
                print('this is blob: ', blob)

                blob.upload_from_filename(local_file)
                print('done upload file: ', local_file)

        print("files uploaded to",
              f'https://console.cloud.google.com/storage/browser/{bucket_name}/{folder_path}'
              f'{local_path}?project={project_id}')
        print("to run tensorboard locally run")
        print(f'1. gsutil cp -r gs://{bucket_name}/{folder_path}{local_path}/ ./trainer/training_output')
        print('2. tensorboard --logdir=./trainer/training_output/')


##
# @file
#
# @brief Main usage simulation.


if __name__ == '__main__':
    ## Main for training process
    main_class = Main(cfg=config, training_path=None, validation_path=None)  # Example: training_path=os.getcwd() + '/datasets')
    main_class.execute_trainer()

    ## Path to last saved model
    current_folder = os.path.dirname(os.path.realpath(__file__))
    last_model_path = os.path.join(current_folder, 'training_output')
    last_model_path = sorted(Path(last_model_path).rglob('*.ckpt'), key=os.path.getmtime, reverse=True)[0]

    ## Main for testing process
    main_class_test = Main(cfg=config, test_dataset_path=None)  # Example: test_dataset_path=os.getcwd() + '/datasets'
    test_results = main_class_test.execute_test(model_path=last_model_path)
    print('*** test results:\n', test_results)

    ## Main for evaluation process
    main_class_eval = Main(cfg=config, test_dataset_path=None)  # Example: test_dataset_path=os.getcwd() + '/datasets'
    eval_results = main_class_eval.execute_eval()
    print('*** Eval results:\n', eval_results)

    ## The results of the comparison between trained model and existing model evaluation.
    comparison_results = generatedProjectNameComparison(
        test_results=test_results,
        eval_results=eval_results,
        metrics_config=config['metrics']).compare()

    print('comparison_results', comparison_results)

    print("Done !!!")



from dsframework.base.trainer.comparison import ZIDSTrainerComparison

##
# @file
# @brief generatedClass class, executes metrics results comparison for two models.


class generatedClass(ZIDSTrainerComparison):
    """ Comparison class, is where we compare two model test metric results.

    Basic comparison is supplied in base class of the compare() method. Override compare as required.
    Metrics evaluated are defined in config.py -> metrics section.
    """
    comparison_results = {}

    def __init__(self, test_results, eval_results, metrics_config):
        super().__init__(test_results, eval_results, metrics_config)

    def compare(self):
        """! Implement you comparison between your model and the evaluated model.
        Basic comparison is supplied in base class of the compare() method.

        For example:
            @code
            if self.test_results is None or self.eval_results is None:
                return "No test results found."

            self.comparison_results = {
                metric: {'trained': self.test_results[metric], 'eval': self.eval_results[metric]}
                for metric in self.metrics
                if metric in self.test_results and metric in self.eval_results and self.metrics[metric]}

            return self.comparison_results
            @endcode
        """

        if not self.comparison_results:
            raise "Compare not implemented. Please implement in comparison.py -> compare()"



"""! @brief Pipeline test entry point."""
from pipeline.pipeline import generatedProjectNamePipeline
from pipeline.schema.inputs import generatedProjectNameInputs

##
# @file
#
# @brief Pipeline test entry point.
if __name__ == '__main__':
    ## Creates an instance of the pipeline
    p = generatedProjectNamePipeline()

    ## This method executes the pipeline, dataset and required params can be loaded here.\n
    # Examples:
    # @code
    # data = {}
    # for file in data_files:
    #     with open(file) as f:
    #         data = load_json(f)
    #
    # output = p.execute(**data)
    # @endcode
    # or
    # @code
    # p.execute(data=mydata)
    # @endcode
    # or
    # @code
    # sig = Signature(uid=1, text=signature, hints=None)
    # output = p.execute(signatures=[sig])
    # @endcode
    # Define data passed through the execute method in the schema/<my-new-project>Inputs.py class, for example:
    # @code
    # class <my-new-project>Inputs(BaseModel):
    #     signatures: List[Signature]
    # @endcode
    # or
    # @code
    # class <my-new-project>Inputs(BaseModel):
    #     data = {}
    # @endcode

    data = {}
    d = generatedProjectNameInputs.parse_obj(data)

    output = p.execute(**d.dict())
    print(output)



#!/usr/bin/env python
# coding: utf-8

from pydantic.main import BaseModel

##
# @file
# @brief Setting the schema inputs, which is the basic structure that the pipeline/service would receive.
#        What data variables you need as an input? it could be a dataset or part of it and additional
#        required variables, such as threshold etc.
#        Using Pydantic's base class that can validate their input class.


class generatedClass(BaseModel):
    """! For example :
    @code
    class <my-new-project>Inputs(BaseModel):
       email: str = ""
       threshold: Optional[float]
       hints: Optional[List[EmailHint]]
    @endcode
    Another example:
    @code
    class <my-new-project>Inputs(BaseModel):
        html_content: str
        source: str
        queue_name: str
        title: str
    @endcode
    """
    pass



from dsframework.base.trainer.pl_wrapper import *

##
# @file
# @brief generatedClass class, pytorch lightning 'Dataset' base class. Implement for external datasets.
class generatedClass(ZIDS_PLCustomDataset):
    """! Template for loading an external custom dataset

    Pytorch lightning docs:
    https://pytorch.org/docs/stable/data.html#torch.utils.data.Dataset
    https://pytorch-lightning.readthedocs.io/en/stable/notebooks/course_UvA-DL/01-introduction-to-pytorch.html?highlight=__getitem__#The-dataset-classs

    """
    def __init__(self, x_ds, y_ds):
        """! Initializer for a custom dataset creation

            Args:
                x_ds: Dataset
                y_ds: Labels for dataset.

            Important note:
                In ZIDSCustomDataset base class there is an implementation of __getitem__ and __len__, which
                are essential for creating a custom Dataset.

        """
        super().__init__(x_ds, y_ds)

    def __len__(self):
        return len(self.y_data)

    def __getitem__(self, index):
        return self.x_data[index], self.y_data[index]



from trainer.pl_wrapper.data_module import generatedProjectNameDataModule
from dsframework.base.trainer.pl_wrapper import *
from dsframework.base.trainer.data import ZIDSTrainerData

##
# @file
# @brief generatedClass class, Datasets and functionality related to datasets.


class generatedClass(ZIDSTrainerData):
    """generatedClass, holds datasets and functionality related to it."""
    dataset_train = None
    dataset_validation = None
    dataset_test = None
    data_module = None

    def __init__(self, model_config=None, load_datasets=False, training_path=None, validation_path=None,
                 test_dataset_path=None, tokenizer=None):
        """generatedClass class initializer."""

        super().__init__(model_config, load_datasets, training_path, validation_path, test_dataset_path, tokenizer)

    def create_data_module(self):
        """! Creates data module in the format you use in the fit/train method. Called from super().__init__().

        Needs to be implemented.

        Example (pytorch lightning):
            self.data_module = generatedProjectNameDataModule(
                model_config=self.model_config,
                train_set=self.dataset_train,
                val_set=self.dataset_validation,
                test_set=self.dataset_test
            )

            return self.data_module
        """
        if self.data_module is None:
            raise Exception('Implement data module in data.py -> create_data_module()')

    def get_data_module(self):
        """! Returns data module in the format you use in the fit/train method.

        Returns:
            data_module

        """
        if self.data_module is None:
            raise Exception('Implement data module in data.py -> create_data_module()')

        return self.data_module

    def load_datasets(self, ds_training_path=None, ds_val_path=None, ds_test_path=None):
        """! Dataset loading.

        Args:
            ds_training_path: Path to training datasets
            ds_val_path: Path to validation datasets
            ds_test_path: Path to test datasets


        Example:
            @code
            if ds_training_path is None and ds_test_path is not None:
                ds_training_path = ds_test_path

            dataset = datasets.MNIST(ds_training_path, download=True, train=True,
                                     transform=transforms.ToTensor())
            self.dataset_train, self.dataset_validation = random_split(dataset, [55000, 5000])
            self.dataset_test = datasets.MNIST(ds_training_path, download=True, train=False,
                                               transform=transforms.ToTensor())

            if self.model_config['num_classes_conf_mat'] is None:
                try:
                    self.model_config['num_classes_conf_mat'] = len(self.dataset_train.dataset.classes)
                except Exception as e:
                    self.model_config['num_classes_conf_mat'] = None
                    print(f'Exception getting number of classes for confusion matrix, please set manually in config.py: {e}')
            @endcode
        """
        raise Exception('load_datasets() not implemented.')



from dataclasses import dataclass
from enum import Enum
from typing import Any, List


@dataclass
class CacheItem:
    """!Data class for passing (key,value) pairs into the cache"""
    key: str
    value: Any


class Cache:
    """!Generic cache interface. Does not contain any implementation."""

    class Type(Enum):
        """!Cache types Enum - NONE(0), REDIS(1), IN_MEMORY(2)"""
        NONE = 0
        REDIS = 1
        IN_MEMORY = 2  # for testing purposes only

    def init(self):
        """!Initialization for the cache in any manner (for example: connect)"""
        pass

    def get(self, keys: List[str]) -> List[Any]:
        """!Retrieve one or more items from the cache.

        :param keys: the list of keys to retrieve items for
        :return: a list of one or more items corresponding to the given list of keys
        """
        raise NotImplementedError

    def set(self, cache_items: List[CacheItem]) -> None:
        """!Store one or more items in the cache.

        :param cache_items: the list of CacheItems (key:value pairs) to be stored
        """
        raise NotImplementedError

    def status(self) -> bool:
        """!Indicate if the cache client is ready for work or not.

        :return: True if ready for work, False otherwise
        """
        raise NotImplementedError



from itertools import cycle
from dsframework.base.trainer.pl_wrapper import *

##
# @file
# @brief generatedClass class, pytorch lightning 'IterableDataset' base class. Implement for external datasets.
class generatedClass(ZIDSIterableDataset):
    """! This is the template class for iterable datasets,
    It basically returns the dataset records by batch.

    Pytorch lightning docs:
    https://pytorch-lightning.readthedocs.io/en/latest/guides/data.html?highlight=IterableDataset
    https://pytorch-lightning.readthedocs.io/en/latest/guides/data.html?highlight=IterableDataset#iterable-datasets
    https://pytorch-lightning.readthedocs.io/en/latest/clouds/fault_tolerant_training_faq.html?highlight=IterableDataset#how-do-i-use-iterable-datasets
    https://pytorch.org/docs/stable/data.html#torch.utils.data.IterableDataset

    Usage:
        override __iter__ method.

    """

    def __init__(self, data):
        self.data_source = data

    def process_data(self, data):
        """Yields a 'rec' out of the dataset."""
        for rec in data:
            yield rec

    def get_stream(self, data):
        """Cycles through dataset and calls process_data method.
        Sending batches can be implemented here."""
        return cycle(self.process_data(data))

    def __iter__(self):
        return self.get_stream(self.data_source)


def worker_init_fn():
    worker_info = get_worker_info()
    dataset = worker_info.dataset  # the dataset copy in this worker process
    worker_id = worker_info.id
    split_size = len(dataset.data) // worker_info.num_workers
    dataset.data = dataset.data[worker_id * split_size:(worker_id + 1) * split_size]



import requests
import json
import sys
from server.token_generator import tokenGenerator

env = sys.argv[1]

t = tokenGenerator()
jwtToken = t.generateToken(env.upper() + '_DSP_Conf')

main_model = 'original_project_name'
repo = 'dozi/' + main_model
data = {}
url = "https://dsp.zoominfo.com/api/v1/models/getModelOptions"
if env == 'stg':
	url = "https://dsp-stg.zoominfo.com/api/v1/models/getModelOptions"
headers = {
	'x-token': 'fake-super-secret-token',
	'Authorization': 'Bearer ' + jwtToken
}

response = requests.post(url, json=data, headers=headers)
try:
	result = json.loads(response.content)
	if 'data' in result:
		models = result['data']['models']
		model_options = [{'name': model['name'], 'id': model['id'], 'repo': model['repo']} for model in models if model['repo'] == repo]
		if len(model_options) == 0:
			print(f'no model named {main_model} in dsp')
		else:
			if not len(model_options):
				print(f'cant find model name {main_model} in dsp')
			else:
				print(model_options)
	else:
		print(f'cant find model name {main_model} in dsp')
except Exception as e:
	print('error', e)
